"""
Monitors for SimpleMonitor
"""

from .arlo import MonitorArloCamera
from .compound import CompoundMonitor, RemoteHostsMonitor
from .file import MonitorBackup
from .gmirror import MonitorGmirrorStatus
from .hass import MonitorSensor
from .host import (
    MonitorApcupsd,
    MonitorCommand,
    MonitorDiskSpace,
    MonitorFileStat,
    MonitorLoadAvg,
    MonitorMemory,
    MonitorNUT,
    MonitorPkgAudit,
    MonitorPortAudit,
    MonitorSwap,
    MonitorZap,
)
from .network import (
    MonitorDNS,
    MonitorHost,
    MonitorHTTP,
    MonitorPing,
    MonitorTCP,
    MonitorTLSCert,
)
from .ring import MonitorRingDoorbell
from .service import (
    MonitorEximQueue,
    MonitorProcess,
    MonitorRC,
    MonitorService,
    MonitorSvc,
    MonitorSystemdUnit,
    MonitorUnixService,
    MonitorWindowsDHCPScope,
)
from .unifi import MonitorUnifiFailover, MonitorUnifiFailoverWatchdog

__all__ = [
    "CompoundMonitor",
    "MonitorApcupsd",
    "MonitorArloCamera",
    "MonitorBackup",
    "MonitorCommand",
    "MonitorDNS",
    "MonitorDiskSpace",
    "MonitorEximQueue",
    "MonitorFileStat",
    "MonitorGmirrorStatus",
    "MonitorHTTP",
    "MonitorHost",
    "MonitorLoadAvg",
    "MonitorMemory",
    "MonitorNUT",
    "MonitorPing",
    "MonitorPkgAudit",
    "MonitorPortAudit",
    "MonitorProcess",
    "MonitorRC",
    "MonitorRingDoorbell",
    "MonitorSensor",
    "MonitorService",
    "MonitorSvc",
    "MonitorSwap",
    "MonitorSystemdUnit",
    "MonitorTCP",
    "MonitorTLSCert",
    "MonitorUnifiFailover",
    "MonitorUnifiFailoverWatchdog",
    "MonitorUnixService",
    "MonitorWindowsDHCPScope",
    "MonitorZap",
    "RemoteHostsMonitor",
]



"""
Network-related monitors for SimpleMonitor
"""

import datetime
import json
import re
import socket
import ssl
import subprocess  # nosec
import sys
from typing import List, Optional, Pattern, Tuple, Union, cast

import arrow
import requests
from requests.auth import HTTPBasicAuth

from .monitor import Monitor, register

try:
    from icmplib import NameLookupError, SocketPermissionError, ping

except ImportError:
    pass


@register
class MonitorHTTP(Monitor):
    """Check an HTTP server is working right."""

    url = ""
    regexp = None
    regexp_text = ""
    allowed_codes = []  # type: List[int]
    allow_redirects = True  # type: bool

    monitor_type = "http"
    method = "GET"

    # optional - for HTTPS client authentication only
    certfile: Optional[str] = None
    keyfile: Optional[str] = None
    cert: Optional[tuple] = None

    auth: Optional[HTTPBasicAuth] = None

    # optional - headers
    headers: Optional[dict] = None
    data: Optional[str] = None

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        self.url = self.get_config_option("url", required=True)

        method = self.get_config_option("method")
        if method is not None:
            method = method.strip().upper()
            if method in ["GET", "POST", "HEAD"]:
                self.method = method
            else:
                raise ValueError("unsupported HTTP method")

        regexp = self.get_config_option("regexp")
        if regexp is not None:
            self.regexp = re.compile(regexp)
            self.regexp_text = regexp

        self.allowed_codes = self.get_config_option(
            "allowed_codes", default=[200], required_type="[int]"
        )
        self.allow_redirects = self.get_config_option(
            "allow_redirects",
            default=True,
            required_type="bool",
        )

        # optional - for HTTPS client authentication only
        # in this case, certfile is required
        self.certfile = cast(Optional[str], config_options.get("certfile"))
        self.keyfile = cast(Optional[str], config_options.get("keyfile"))
        if self.certfile and not self.keyfile:
            self.keyfile = self.certfile
        if not self.certfile and self.keyfile:
            raise ValueError("config option keyfile is set but certfile is not")

        if self.certfile is not None and self.keyfile is not None:
            self.cert = (self.certfile, self.keyfile)
            assert self.certfile is not None and self.keyfile is not None
        else:
            self.cert = None

        headers = config_options.get("headers")
        if headers:
            try:
                self.headers = json.loads(headers)
            except json.JSONDecodeError as e:
                self.monitor_logger.error(f"Parsing headers to JSON failed: {e}")
                self.headers = None

        self.data = None
        self.json = None

        if self.method == "POST":
            data_json = config_options.get("json")
            if data_json:
                try:
                    self.json = json.loads(data_json)
                except json.JSONDecodeError as e:
                    self.monitor_logger.error(f"Parsing json to JSON failed: {e}")
                    self.json = None

            self.data = config_options.get("data")

        if self.json and self.data:
            self.data = None
            raise ValueError("Use only one option - either json OR data")

        self.verify_hostname = self.get_config_option(
            "verify_hostname", default=True, required_type="bool"
        )

        self.request_timeout = self.get_config_option(
            "timeout", default=5, required_type="int"
        )

        self.username = cast(Optional[str], config_options.get("username"))
        self.password = cast(Optional[str], config_options.get("password"))

        if self.username is not None and self.password is not None:
            self.auth = HTTPBasicAuth(self.username, self.password)
        else:
            self.auth = None

    def run_test(self) -> bool:
        start_time = arrow.get()
        end_time = None
        try:
            response = requests.request(
                self.method,
                self.url,
                headers=self.headers,
                timeout=self.request_timeout,
                verify=self.verify_hostname,
                allow_redirects=self.allow_redirects,
                auth=self.auth,
                cert=self.cert,
                data=self.data,
                json=self.json,
            )
            end_time = arrow.get()
            load_time = end_time - start_time
            if response.status_code not in self.allowed_codes:
                return self.record_fail(
                    "Got status '{0} {1}' instead of {2}".format(
                        response.status_code, response.reason, self.allowed_codes
                    )
                )
            if self.regexp is None:
                return self.record_success(
                    "%s in %0.2fs"
                    % (
                        response.status_code,
                        (load_time.seconds + (load_time.microseconds / 1000000.2)),
                    )
                )
            matches = self.regexp.search(response.text)
            if matches:
                return self.record_success(
                    "%s in %0.2fs"
                    % (
                        response.status_code,
                        (load_time.seconds + (load_time.microseconds / 1000000.2)),
                    )
                )
            return self.record_fail(
                "Got '{0} {1}' but couldn't match /{2}/ in page.".format(
                    response.status_code, response.reason, self.regexp_text
                )
            )
        except requests.exceptions.SSLError:
            return self.record_fail("SSL error during connection")
        except requests.exceptions.RequestException as exception:
            return self.record_fail(
                "Requests exception while opening URL: {0}".format(exception)
            )

    def describe(self) -> str:
        """Explains what we do."""
        codes = [str(x) for x in self.allowed_codes]
        message = "Checking {} returns HTTP/{} within {}s".format(
            self.url, "or".join(codes), self.request_timeout
        )
        if self.regexp is not None:
            message = message + " and that /{}/ matches the page".format(
                self.regexp_text
            )
        return message

    def get_params(self) -> Tuple:
        return (self.url, self.regexp_text, self.allowed_codes)


@register
class MonitorTCP(Monitor):
    """TCP port monitor"""

    host = ""
    port = 0
    monitor_type = "tcp"

    def __init__(self, name: str, config_options: dict) -> None:
        """Constructor"""
        super().__init__(name, config_options)
        self.host = self.get_config_option("host", required=True)
        self.port = cast(
            int,
            self.get_config_option(
                "port", required=True, required_type="int", minimum=0
            ),
        )

    def run_test(self) -> bool:
        """Check the port is open on the remote host"""
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.settimeout(5.0)
            sock.connect((self.host, self.port))
        except OSError as exception:
            return self.record_fail(str(exception))
        sock.close()
        return self.record_success()

    def describe(self) -> str:
        """Explains what this instance is checking"""
        return "checking for open tcp socket on %s:%d" % (self.host, self.port)

    def get_params(self) -> Tuple:
        return (self.host, self.port)


@register
class MonitorHost(Monitor):
    """Ping a host to make sure it's up"""

    host = ""
    ping_command = ""
    ping_regexp = ""
    monitor_type = "host"
    time_regexp = ""
    r = ""  # type: Union[str, Pattern[str]]
    r2 = ""  # type: Union[str, Pattern[str]]

    def __init__(self, name: str, config_options: dict) -> None:
        """
        Note: We use -w/-t on Windows/POSIX to limit the amount of time we wait to 5 seconds.
        This is to stop ping holding things up too much. A machine that can't ping back in <5s is
        a machine in trouble anyway, so should probably count as a failure.
        """
        super().__init__(name, config_options)
        ping_ttl = self.get_config_option(
            "ping_ttl", required_type="int", minimum=0, default=5
        )
        ping_ms = str(ping_ttl * 1000)
        ping_ttl = str(ping_ttl)
        self.count = cast(
            int,
            self.get_config_option("count", required_type="int", default=1, minimum=1),
        )
        platform = sys.platform
        if platform in ["win32", "cygwin"]:
            self.ping_command = f"ping -n {self.count} -w " + ping_ms + " %s"
            self.ping_regexp = r"Reply from [0-9a-f:.]+:.+time[=<]\d+ms"
            self.time_regexp = r"Average = (?P<ms>\d+)ms"
        elif platform.startswith("freebsd") or platform.startswith("darwin"):
            self.ping_command = f"ping -c{self.count} -t" + ping_ttl + " %s"
            self.ping_regexp = "bytes from"
            self.time_regexp = r"min/avg/max/stddev = [\d.]+/(?P<ms>[\d.]+)/"
        elif platform.startswith("linux"):
            self.ping_command = f"ping -c{self.count} -W" + ping_ttl + " %s"
            self.ping_regexp = "bytes from"
            self.time_regexp = r"min/avg/max/stddev = [\d.]+/(?P<ms>[\d.]+)/"
        else:
            RuntimeError("Don't know how to run ping on this platform, help!")
        self.ping_regexp = self.get_config_option(
            "ping_regexp", required=False, default=self.ping_regexp
        )
        self.time_regexp = self.get_config_option(
            "time_regexp", required=False, default=self.ping_regexp
        )

        self.host = self.get_config_option("host", required=True)

    def run_test(self) -> bool:
        success = False
        pingtime = 0.0

        try:
            cmd = (self.ping_command % self.host).split(" ")
            output = subprocess.check_output(cmd)  # nosec
            for line in str(output).split("\n"):
                matches = re.search(self.ping_regexp, line)
                if matches:
                    success = True
                else:
                    matches = re.search(self.time_regexp, line)
                    if matches:
                        pingtime = float(matches.group("ms"))
        except subprocess.CalledProcessError as exception:
            return self.record_fail(str(exception))
        if success:
            if pingtime > 0:
                return self.record_success("%sms" % pingtime)
            return self.record_success()
        return self.record_fail()

    def describe(self) -> str:
        """Explains what this instance is checking"""
        return "checking host %s is pingable" % self.host

    def get_params(self) -> Tuple:
        return (self.host,)


@register
class MonitorDNS(Monitor):
    """Monitor DNS server."""

    monitor_type = "dns"
    path = ""
    command = "dig"

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        self.path = self.get_config_option("record", required=True)
        self.desired_val = self.get_config_option("desired_val")
        self.server = self.get_config_option("server")
        self.params = [self.command]
        self.port = self.get_config_option("port", required_type="int")

        if self.server:
            self.params.append("@%s" % self.server)

        self.rectype = self.get_config_option("record_type")
        if self.rectype:
            self.params.append("-t")
            self.params.append(config_options["record_type"])

        self.params.append(self.path)
        self.params.append("+short")

        if self.port:
            self.params.extend(["-p", str(self.port)])

    def run_test(self) -> bool:
        try:
            result = subprocess.check_output(self.params).decode("utf-8")  # nosec
            result = result.strip()
            if result is None or result == "":
                if self.desired_val != "nxdomain":
                    return self.record_fail("failed to resolve %s" % self.path)
                return self.record_success("successfully did not resolve")
            if self.desired_val and set(result.split("\n")) != set(
                self.desired_val.split("\n")
            ):
                return self.record_fail(
                    "resolved DNS record is unexpected: %s != %s"
                    % (self.desired_val, result)
                )
            return self.record_success()
        except subprocess.CalledProcessError as exception:
            return self.record_fail(
                "Command '%s' exited non-zero (%d)"
                % (" ".join(self.params), exception.returncode)
            )

    def describe(self) -> str:
        if self.desired_val:
            end_part = "resolves to %s" % self.desired_val
        else:
            end_part = "is resolvable"

        if self.rectype:
            mid_part = "%s record %s" % (self.rectype, self.path)
        else:
            mid_part = "record %s" % self.path

        if self.server:
            very_end_part = " at %s" % self.server
        else:
            very_end_part = ""
        try:
            port_part = f" on port {self.port}" if self.port else ""
        except AttributeError:
            port_part = ""
        return "Checking that DNS %s %s%s%s" % (
            mid_part,
            end_part,
            very_end_part,
            port_part,
        )

    def get_params(self) -> Tuple:
        return (self.path,)


@register
class MonitorPing(Monitor):
    """Ping a host to make sure it's up, using native Python"""

    monitor_type = "ping"

    def __init__(self, name: str, config_options: dict) -> None:
        if config_options is None:
            config_options = {}
        super().__init__(name=name, config_options=config_options)
        self.host = cast(str, self.get_config_option("host", required=True))
        self.timeout = cast(
            int, self.get_config_option("timeout", required_type="int", default=5)
        )
        self.count = cast(
            int,
            self.get_config_option("count", required_type="int", default=1, minimum=1),
        )

    def run_test(self) -> bool:
        if "icmplib" not in sys.modules:
            return self.record_fail("Missing required icmplib module")
        try:
            result = ping(self.host, count=self.count, timeout=self.timeout)
            if result.is_alive:
                return self.record_success(
                    "RTT for {}: {:0.3f}ms".format(result.address, result.avg_rtt)
                )
            return self.record_fail(f"Host {result.address} is not alive")
        except NameLookupError:
            return self.record_fail(f"Failed to resolve {self.host}")
        except SocketPermissionError:
            return self.record_fail(
                "ping monitor requires root to work; "
                "try the 'host' monitor if this is not an option for you"
            )

    def get_params(self) -> Tuple:
        return (self.host, self.timeout, self.count)

    def describe(self) -> str:
        return "Checking {} pings within {} seconds ({} attempt(s))".format(
            self.host, self.timeout, self.count
        )


@register
class MonitorTLSCert(Monitor):
    """Check the cert on a TLS connection is not due to expire."""

    monitor_type = "tls_expiry"

    def __init__(self, name: str, config_options: Optional[dict]) -> None:
        if config_options is None:
            config_options = {}
        self.minimum_gap = 43200  # 12 hours
        super().__init__(name=name, config_options=config_options)
        self.host = cast(str, self.get_config_option("host", required=True))
        self.port = cast(
            int, self.get_config_option("port", required_type="int", default=443)
        )
        self.min_days = cast(
            int, self.get_config_option("min_days", required_type="int", default=7)
        )
        if self.min_days < 0:
            raise ValueError("min_days must be 0 or greater")
        self.sni = cast(Optional[str], self.get_config_option("sni", required=False))

    def run_test(self) -> bool:
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        ssl_context.verify_mode = ssl.CERT_REQUIRED
        ssl_context.check_hostname = bool(self.sni)
        ssl_context.load_default_certs()

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            with ssl_context.wrap_socket(
                sock, server_hostname=self.sni if self.sni else None
            ) as ssl_sock:
                try:
                    ssl_sock.connect((self.host, self.port))
                except socket.gaierror as error:
                    self.monitor_logger.exception("Failed to connect socket")
                    return self.record_fail("Failed to connect: {}".format(error))
                except ssl.CertificateError as error:
                    self.monitor_logger.exception(
                        "SSL certification validation error: %s", error.verify_message
                    )
                    return self.record_fail(
                        "SSL validation error: {}".format(error.verify_message)
                    )
                except ssl.SSLError as error:
                    self.monitor_logger.exception("SSL Error: %s", error.reason)
                    return self.record_fail("SSL Error: {}".format(error.reason))
                cert = ssl_sock.getpeercert()
                if not cert:
                    return self.record_fail("Did not receive certificate")
                not_after = str(cert["notAfter"])
                expiry = datetime.datetime.strptime(not_after, r"%b %d %H:%M:%S %Y %Z")
                delta = expiry - datetime.datetime.utcnow()
                days_left = delta.days
                if days_left < self.min_days:
                    if days_left < 0:
                        return self.record_fail(
                            "Certificate at {}:{} expired {} days ago".format(
                                self.host, self.port, abs(days_left)
                            )
                        )
                    return self.record_fail(
                        "Certificate at {}:{} expires in {} days".format(
                            self.host, self.port, days_left
                        )
                    )
                return self.record_success(
                    "Certificate at {}:{} has {} days left to expiry".format(
                        self.host, self.port, days_left
                    )
                )

    def get_params(self) -> Tuple:
        return (self.host, self.port, self.min_days)

    def describe(self) -> str:
        return "Checking TLS cert at {}:{} {}has at least {} days until expiry".format(
            self.host,
            self.port,
            "(sni: " + self.sni + ") " if self.sni else "",
            self.min_days,
        )



"""
SimpleMonitor logging to files
"""

import glob
import json
import logging
import logging.handlers
import os
import shutil
import socket
import stat
import subprocess  # nosec
import sys
import tempfile
from typing import TYPE_CHECKING, Any, Dict, List, Optional, cast

import arrow
from jinja2 import Environment, FileSystemLoader, select_autoescape

from ..Monitors.monitor import Monitor
from ..util import (
    copy_if_different,
    format_datetime,
    short_hostname,
    size_string_to_bytes,
)
from ..version import VERSION
from .logger import Logger, register

if TYPE_CHECKING:
    import datetime


@register
class FileLogger(Logger):
    """Log monitor status to a file."""

    logger_type = "logfile"
    filename = ""
    only_failures = False
    buffered = True
    dateformat = None

    def __init__(self, config_options: Optional[Dict[str, Any]] = None) -> None:
        if config_options is None:
            config_options = {}
        super().__init__(config_options)
        self.filename = self.get_config_option(
            "filename", required=True, allow_empty=False
        )
        self.file_handle = open(self.filename, "a+")  # pylint: disable=consider-using-with

        self.only_failures = self.get_config_option(
            "only_failures", required_type="bool", default=False
        )

        self.buffered = self.get_config_option(
            "buffered", required_type="bool", default=True
        )

        self.file_handle.write(
            "{} simplemonitor starting\n".format(self._get_datestring())
        )
        if not self.buffered:
            self.file_handle.flush()

    def __del__(self) -> None:
        self.file_handle.close()

    def save_result2(self, name: str, monitor: Monitor) -> None:
        if self.only_failures and monitor.virtual_fail_count() == 0:
            return

        try:
            if monitor.virtual_fail_count() > 0:
                self.file_handle.write(
                    "%s %s: failed since %s; VFC=%d (%s) (%0.3fs)"
                    % (
                        self._get_datestring(),
                        name,
                        format_datetime(monitor.first_failure_time(), self.tz),
                        monitor.virtual_fail_count(),
                        monitor.get_result(),
                        monitor.last_run_duration,
                    )
                )
            else:
                self.file_handle.write(
                    "%s %s: ok (%0.3fs)"
                    % (self._get_datestring(), name, monitor.last_run_duration)
                )
            self.file_handle.write("\n")

            if not self.buffered:
                self.file_handle.flush()
        except OSError:
            self.logger_logger.exception("Error writing to logfile %s", self.filename)

    def hup(self) -> None:
        """Close and reopen log file."""
        try:
            self.file_handle.close()
            self.file_handle = open(self.filename, "a+")  # pylint: disable=consider-using-with
        except OSError:
            self.logger_logger.exception(
                "Couldn't reopen log file %s after HUP", self.filename
            )

    def describe(self) -> str:
        return "Writing log file to {0}".format(self.filename)


@register
class FileLoggerNG(Logger):
    """
    Log monitor status to a file, Next Generation

    Uses the Python logging library to get features like rotation
    """

    logger_type = "logfileng"
    only_failures = False

    def __init__(self, config_options: Optional[dict] = None) -> None:
        if config_options is None:
            config_options = {}
        super().__init__(config_options)
        self.only_failures = self.get_config_option(
            "only_failures", required_type="bool", default=False
        )
        self._logger = logging.getLogger(f"logfileng-{self.name}")
        if self.only_failures:
            self._logger.setLevel(logging.WARNING)
        else:
            self._logger.setLevel(logging.INFO)
        rotation_type = self.get_config_option(
            "rotation_type", allowed_values=["time", "size"]
        )
        self.filename = self.get_config_option("filename")
        self.backup_count = cast(
            int, self.get_config_option("backup_count", required_type="int", default=1)
        )
        if rotation_type == "time":
            handler = logging.handlers.TimedRotatingFileHandler(
                filename=self.filename,
                when=self.get_config_option("when", default="h"),
                interval=self.get_config_option(
                    "interval", default=1, required_type="int"
                ),
                backupCount=self.backup_count,
                utc=self.get_config_option("utc", required_type="bool", default=True),
                encoding="utf-8",
            )  # type: logging.handlers.BaseRotatingHandler
        elif rotation_type == "size":
            max_bytes = size_string_to_bytes(self.get_config_option("max_bytes"))
            if max_bytes is None:
                raise ValueError("Missing max_bytes")
            handler = logging.handlers.RotatingFileHandler(
                filename=self.filename,
                maxBytes=max_bytes,
                backupCount=self.backup_count,
                encoding="utf-8",
            )
        else:
            raise ValueError(f"Invalid rotation_type {rotation_type}")
        self._logger.addHandler(handler)

    def save_result2(self, name: str, monitor: Monitor) -> None:
        if monitor.virtual_fail_count() == 0:
            self._logger.info(
                "%s %s: ok (%0.3fs) (%s)",
                self._get_datestring(),
                name,
                monitor.last_run_duration,
                monitor.get_result(),
            )
        else:
            self._logger.warning(
                "%s %s: failed since %s; VFC=%d (%s) (%0.3fs)",
                self._get_datestring(),
                name,
                format_datetime(monitor.first_failure_time(), self.tz),
                monitor.virtual_fail_count(),
                monitor.get_result(),
                monitor.last_run_duration,
            )


@register
class HTMLLogger(Logger):
    """A batching logger which writes a simple HTML page of the current state."""

    logger_type = "html"
    supports_batch = True
    filename = ""
    count_data = ""

    def __init__(self, config_options: Optional[dict] = None) -> None:
        if config_options is None:
            config_options = {}
        super().__init__(config_options)
        package_path = sys.modules["simplemonitor"].__file__ or "."
        package_data_dir = os.path.join(os.path.dirname(package_path), "html")
        self.filename = cast(
            str, self.get_config_option("filename", required=True, allow_empty=False)
        )
        self.source_folder = cast(
            str,
            self.get_config_option(
                "source_folder", required=False, default=package_data_dir
            ),
        )
        self.folder = cast(
            str, self.get_config_option("folder", required=False, default="html")
        )
        if not os.path.isdir(self.folder):
            self.logger_logger.critical("output folder %s does not exist", self.folder)
        self.copy_resources = cast(
            bool,
            self.get_config_option(
                "copy_resources", required_type="bool", default=True
            ),
        )
        self.navbar_title = cast(
            str,
            self.get_config_option(
                "navbar_title", required=False, default="SimpleMonitor"
            ),
        )
        self.visual_theme = cast(
            str,
            self.get_config_option("visual_theme", required=False, default="light"),
        )
        self.upload_command = cast(
            str,
            self.get_config_option("upload_command", required=False, allow_empty=False),
        )
        self.map = cast(
            bool, self.get_config_option("map", required_type="bool", default=False)
        )
        _map_start = cast(
            Optional[List[str]],
            self.get_config_option("map_start", required_type="[str]", default=None),
        )
        if _map_start and len(_map_start) == 3:
            self.map_start = {
                "latitude": _map_start[0],
                "longitude": _map_start[1],
                "zoom": _map_start[2],
            }  # type: Optional[Dict[str, str]]
        else:
            self.map_start = None
        if self.map and not self.map_start:
            raise RuntimeError(
                f"map is set for logger {self.name} but map_start is missing or badly formatted"
            )
        self.map_token = cast(str, self.get_config_option("map_token"))
        if self.map_token:
            self.logger_logger.info(
                "map_token option for logger %s is no longer required; ignoring",
                self.name,
            )
        self._resource_files = [
            "dist/main.bundle.js*",
            "dist/maps.bundle.js*",
            "dist/*.png",
        ]
        self._my_host = short_hostname()
        self.status = ""
        self.header_class = ""
        self._env = Environment(
            loader=FileSystemLoader(self.source_folder),
            keep_trailing_newline=True,
            autoescape=select_autoescape("html"),
        )

    def save_result2(self, name: str, monitor: Monitor) -> None:
        if not self.doing_batch:
            self.logger_logger.error(
                "HTMLLogger.save_result2() called while not doing batch."
            )
            return
        if self.batch_data is None:
            self.batch_data = {}
        status = bool(monitor.virtual_fail_count() == 0)
        if not status:
            fail_time = format_datetime(monitor.first_failure_time(), self.tz)
            fail_count = monitor.virtual_fail_count()
            fail_data = monitor.get_result()
            downtime = monitor.get_downtime()
        else:
            fail_time = ""
            fail_count = 0
            fail_data = monitor.get_result()
            downtime = monitor.get_uptime()  # yes, I know
        failures = monitor.failures
        last_failure = format_datetime(monitor.last_failure, self.tz)
        gap = monitor.minimum_gap
        if gap == 0:
            # TODO: figure out a good way to know the interval value for both local and
            # remote monitors
            gap = 60

        try:
            if monitor.last_update is not None:
                age = arrow.utcnow() - monitor.last_update  # type: datetime.timedelta
                age_seconds = age.days * 3600 + age.seconds
                update = str(monitor.last_update)
            else:
                raise ValueError
        except ValueError:
            age_seconds = 0
            update = ""
        row_class = ""
        cell_class = ""
        if not monitor.enabled:
            status_text = "DISABLED"
        elif age_seconds > gap + 60:
            status_text = "OLD"
            cell_class = "table-warning"
        elif status:
            status_text = "OK"
            cell_class = "table-success"
        else:
            status_text = "FAIL"
            row_class = "table-danger"

        data_line = {
            "monitor_name": monitor.name,
            "status": status,
            "status_text": status_text,
            "row_class": row_class,
            "cell_class": cell_class,
            "fail_time": fail_time,
            "fail_count": fail_count,
            "fail_data": fail_data,
            "downtime": downtime,
            "age": age_seconds,
            "update": update,
            "host": monitor.running_on,
            "failures": failures,
            "last_failure": last_failure,
            "gap": gap,
            "availability": monitor.availability,
            "description": monitor.describe(),
            "link": monitor.failure_doc,
            "enabled": monitor.enabled,
            "my_host": monitor.running_on == self._my_host,
            "gps": monitor.gps,
        }  # type: Dict[str, Any]
        self.batch_data[monitor.name] = data_line

    def process_batch(self) -> None:
        """Save the HTML file."""
        ok_count = 0
        fail_count = 0
        old_count = 0
        remote_count = 0
        disabled_count = 0

        try:
            temp_file = tempfile.mkstemp()
            file_handle = os.fdopen(temp_file[0], "w")
            file_name = temp_file[1]
        except OSError:
            self.logger_logger.exception(
                "Couldn't create temporary file for HTML output"
            )
            return

        fail_entries = []  # type: List[Dict[str, Any]]
        ok_entries = []  # type: List[Dict[str, Any]]

        if self.batch_data is None:
            return
        keys = list(self.batch_data.keys())
        keys.sort()
        for entry in keys:
            this_entry = self.batch_data[entry]
            this_list = ok_entries
            if not this_entry["enabled"]:
                disabled_count += 1
            elif this_entry["age"] > this_entry["gap"] + 60:
                old_count += 1
            elif this_entry["status"]:
                ok_count += 1
            else:
                fail_count += 1
                this_list = fail_entries
            if this_entry["host"] != self._my_host:
                remote_count += 1
            # output.write(self._make_html_row(entry, this_entry))
            this_list.append(this_entry)
        if old_count > 0:
            self.header_class = "border-warning"
            self.status = "OLD"
        elif fail_count > 0:
            self.header_class = "border-danger"
            self.status = "FAIL"
        else:
            self.header_class = "border-success"
            self.status = "OK"

        template = self._env.get_template("status-template.html")
        if self._global_info:
            interval = max(30, self._global_info["interval"])
        else:
            interval = 30
        file_handle.write(
            template.render(
                navbar_title=self.navbar_title,
                visual_theme=self.visual_theme,
                status=self.status,
                status_border=self.header_class,
                host=socket.gethostname(),
                interval=interval,
                timestamp=str(arrow.now().int_timestamp),
                now=format_datetime(arrow.now(), self.tz),
                version=VERSION,
                ok_count=ok_count,
                fail_count=fail_count,
                disabled_count=disabled_count,
                old_count=old_count,
                remote_count=remote_count,
                fail_entries=fail_entries,
                ok_entries=ok_entries,
                map=self.map,
                map_start=self.map_start,
                map_token=self.map_token,
            )
        )

        try:
            file_handle.flush()
            file_handle.close()
            os.chmod(
                file_name, stat.S_IREAD | stat.S_IWRITE | stat.S_IRGRP | stat.S_IROTH
            )
            if not os.path.isdir(self.folder):
                self.logger_logger.critical(
                    "Target folder %s does not exist", self.folder
                )
                return
            shutil.move(file_name, os.path.join(self.folder, self.filename))
            if self.copy_resources:
                for fileglob in self._resource_files:
                    for filename in glob.glob(
                        os.path.join(self.source_folder, fileglob)
                    ):
                        if copy_if_different(
                            os.path.join(self.source_folder, filename), self.folder
                        ):
                            self.logger_logger.info(f"copied {filename}")
        except OSError:
            self.logger_logger.exception("problem closing/moving files for HTML output")
        if self.upload_command:
            try:
                subprocess.run(self.upload_command.split(" "), check=True)  # nosec
            except subprocess.SubprocessError:
                self.logger_logger.exception(
                    "Failed to run upload command for HTML files"
                )
            except FileNotFoundError:
                self.logger_logger.exception("Could not find upload command")

    def describe(self) -> str:
        return "Writing HTML page to {0}".format(self.filename)


class MonitorResult:
    """Represent the current status of a Monitor."""

    def __init__(self) -> None:
        self.virtual_fail_count = 0
        self.result = None  # type: Optional[str]
        self.first_failure_time = None  # type: Optional[str]
        self.last_run_duration = None  # type: Optional[int]
        self.status = "Fail"
        self.dependencies = []  # type: List[str]

    def json_representation(self) -> dict:
        """Get JSON representation"""
        return self.__dict__


class MonitorJsonPayload:
    def __init__(self) -> None:
        self.generated = None  # type: Optional[str]
        self.monitors = {}  # type: dict

    def json_representation(self) -> dict:
        """Get JSON res""presentation"""
        return self.__dict__


class PayloadEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if hasattr(o, "json_representation"):
            return o.json_representation()
        return json.JSONEncoder.default(self, o.__dict__)


@register
class JsonLogger(Logger):
    """Write monitor status to a JSON file."""

    logger_type = "json"
    filename = ""  # type: str
    supports_batch = True

    def __init__(self, config_options: Optional[dict] = None) -> None:
        if config_options is None:
            config_options = {}
        super().__init__(config_options)
        self.filename = self.get_config_option(
            "filename", required=True, allow_empty=False
        )

    def save_result2(self, name: str, monitor: Monitor) -> None:
        if self.batch_data is None:
            self.batch_data = {}
        result = MonitorResult()
        result.first_failure_time = format_datetime(monitor.first_failure_time())
        result.virtual_fail_count = monitor.virtual_fail_count()
        result.last_run_duration = monitor.last_run_duration
        result.result = monitor.get_result()
        if hasattr(monitor, "was_skipped") and monitor.was_skipped:
            result.status = "Skipped"
        elif monitor.virtual_fail_count() <= 0:
            result.status = "OK"
        result.dependencies = monitor.dependencies

        self.batch_data[name] = result

    def process_batch(self) -> None:
        payload = MonitorJsonPayload()
        payload.generated = format_datetime(arrow.now())
        if self.batch_data is not None:
            payload.monitors = self.batch_data

            with open(self.filename, "w") as outfile:
                json.dump(
                    payload,
                    outfile,
                    indent=4,
                    separators=(",", ":"),
                    ensure_ascii=False,
                    cls=PayloadEncoder,
                )
        self.batch_data = {}

    def describe(self) -> str:
        return "Writing JSON file to {0}".format(self.filename)



"""
SimpleMonitor alerts via ntfy
"""

from typing import cast

import requests

from ..Monitors.monitor import Monitor
from .alerter import Alerter, AlertLength, AlertType, register


@register
class NtfyAlerter(Alerter):
    """Send push notification via ntfy."""

    alerter_type = "ntfy"

    def __init__(self, config_options: dict) -> None:
        super().__init__(config_options)
        self.ntfy_token = cast(str, self.get_config_option("token"))
        self.ntfy_topic = cast(str, self.get_config_option("topic", required=True))
        self.ntfy_server = cast(
            str, self.get_config_option("server", default="https://ntfy.sh")
        )
        self.ntfy_priority = cast(
            str,
            self.get_config_option(
                "priority",
                required_type="str",
                allowed_values=[
                    "max",
                    "urgent",
                    "high",
                    "default",
                    "low",
                    "min",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                ],
                default="default",
            ),
        )
        self.ntfy_tags = cast(str, self.get_config_option("tags", required_type="str"))
        self.timeout = cast(
            int, self.get_config_option("timeout", required_type="int", default=5)
        )
        self.support_catchup = True
        # prefix icon to subject
        self.ntfy_icon_prefix = cast(
            str,
            self.get_config_option("icon_prefix", required_type="bool", default=False),
        )
        self.ntfy_icon_failed = chr(
            int(
                cast(
                    str,
                    self.get_config_option(
                        "icon_failed", required_type="str", default="274C"
                    ),
                ),
                16,
            )
        )
        self.ntfy_icon_succeeded = chr(
            int(
                cast(
                    str,
                    self.get_config_option(
                        "icon_succeeded", required_type="str", default="2705"
                    ),
                ),
                16,
            )
        )

    def send_ntfy_notification(self, subject: str, body: str) -> None:
        """Send a push notification."""
        requests.post(
            f"{self.ntfy_server}/{self.ntfy_topic}",
            data=body,
            headers={
                "Title": subject.encode("UTF-8"),
                "Priority": self.ntfy_priority,
                **({"Tags": self.ntfy_tags} if self.ntfy_tags else {}),
                **(
                    {"Authorization": f"Bearer {self.ntfy_token}"}
                    if self.ntfy_token
                    else {}
                ),
            },
            timeout=self.timeout,
        )

    def send_alert(self, name: str, monitor: Monitor) -> None:
        """Build up the content for the push notification."""

        alert_type = self.should_alert(monitor)

        if alert_type == AlertType.NONE:
            return
        subject = self.build_message(AlertLength.NOTIFICATION, alert_type, monitor)
        body = self.build_message(AlertLength.FULL, alert_type, monitor)

        # prefix icon to subject when relevant
        if self.ntfy_icon_prefix and subject.endswith("failed"):
            subject = f"{self.ntfy_icon_failed} {subject}"
        if self.ntfy_icon_prefix and subject.endswith("succeeded"):
            subject = f"{self.ntfy_icon_succeeded} {subject}"

        if not self._dry_run:
            try:
                self.send_ntfy_notification(subject, body)
            except Exception:
                self.alerter_logger.exception("Couldn't send ntfy notification")
        else:
            self.alerter_logger.info("dry_run: would send nfty notification: %s", body)

    def _describe_action(self) -> str:
        return f"posting to ntfy topic {self.ntfy_topic}"



import datetime
import json
import re
from typing import Any

import arrow

from . import MonitorState

DATETIME_MAGIC_TOKEN = "__simplemonitor_datetime"  # nosec
MONITORSTATE_MAGIC_TOKEN = "__simplemonitor_monitorstate"  # nosec
ARROW_MAGIC_TOKEN = "__simplemonitor_arrow"  # nosec
FORMAT = "%Y-%m-%d %H:%M:%S.%f"


class JSONEncoder(json.JSONEncoder):
    _regexp_type = type(re.compile(""))

    def default(self, o: Any) -> Any:  # pylint: disable=E0202
        if isinstance(o, datetime.datetime):
            return {DATETIME_MAGIC_TOKEN: o.strftime(FORMAT)}
        if isinstance(o, self._regexp_type):
            return "<removed compiled regexp object>"
        if isinstance(o, MonitorState):
            return {MONITORSTATE_MAGIC_TOKEN: o.name}
        if isinstance(o, arrow.Arrow):
            return {ARROW_MAGIC_TOKEN: o.for_json()}
        return super(JSONEncoder, self).default(o)


class JSONDecoder(json.JSONDecoder):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._original_object_pairs_hook = kwargs.pop("object_pairs_hook", None)
        kwargs["object_pairs_hook"] = self.object_pairs_hook
        super(JSONDecoder, self).__init__(*args, **kwargs)

    _datetime_re = re.compile(r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}")

    def object_pairs_hook(self, obj: Any) -> Any:  # pylint: disable=E0202
        if (
            len(obj) == 1
            and obj[0][0] == DATETIME_MAGIC_TOKEN
            and isinstance(obj[0][1], str)
            and self._datetime_re.match(obj[0][1])
        ):
            # convert incoming datetimes to Arrow
            return arrow.get(obj[0][1])
        elif (
            len(obj) == 1
            and obj[0][0] == MONITORSTATE_MAGIC_TOKEN
            and isinstance(obj[0][1], str)
        ):
            return MonitorState[obj[0][1]]
        elif (
            len(obj) == 1
            and obj[0][0] == ARROW_MAGIC_TOKEN
            and isinstance(obj[0][1], str)
        ):
            return arrow.get(obj[0][1])
        elif self._original_object_pairs_hook:
            return self._original_object_pairs_hook(obj)
        else:
            return dict(obj)


def json_dumps(data: Any) -> bytes:
    return JSONEncoder().encode(data).encode("ascii")


def json_loads(string: bytes) -> Any:
    return JSONDecoder().decode(string.decode("ascii"))



"""
SimpleMonitor alerts via Amazon SNS
"""

from typing import Any, Dict, Optional, cast

import boto3
from botocore.exceptions import ClientError

from ..Monitors.monitor import Monitor
from ..util import AlerterConfigurationError
from .alerter import Alerter, AlertLength, AlertType, register


@register
class SNSAlerter(Alerter):
    """Send notifications using Amazon SNS"""

    alerter_type = "sns"
    urgent = True

    def __init__(self, config_options: dict) -> None:
        super().__init__(config_options)
        self.topic = cast(str, self.get_config_option("topic", default=""))
        self.number = cast(str, self.get_config_option("number", default=""))

        if not self.topic and not self.number:
            raise AlerterConfigurationError("need one of topic or number to be set")

        if self.topic and self.number:
            raise AlerterConfigurationError("cannot set both topic and number")

        self.support_catchup = True

        self.sns_client_params: Dict[str, Any] = {}

        aws_region = cast(str, self.get_config_option("aws_region", default=""))
        if aws_region:
            self.sns_client_params["region_name"] = aws_region

        aws_access_key = cast(str, self.get_config_option("aws_access_key", default=""))
        aws_secret_key = cast(
            str, self.get_config_option("aws_secret_access_key", default="")
        )

        if aws_access_key and aws_secret_key:
            self.sns_client_params["aws_access_key_id"] = aws_access_key
            self.sns_client_params["aws_secret_access_key"] = aws_secret_key

        self.sender_id = cast(
            str, self.get_config_option("sender_id", default="SmplMntr")
        )

    def send_alert(self, name: str, monitor: Monitor) -> None:
        """Send the alert"""

        alert_type = self.should_alert(monitor)
        if alert_type == AlertType.NONE:
            return

        subject: Optional[str] = None
        message = "Misconfiguration: could not build message"
        if self.topic:
            subject = self.build_message(AlertLength.NOTIFICATION, alert_type, monitor)
            message = self.build_message(AlertLength.FULL, alert_type, monitor)
        elif self.number:
            message = self.build_message(AlertLength.SMS, alert_type, monitor)

        if not self._dry_run:
            try:
                client = boto3.client("sns", **self.sns_client_params)
                if subject is None:
                    client.publish(
                        PhoneNumber=self.number,
                        Message=message,
                        MessageAttributes={
                            "AWS.SNS.SMS.SenderID": {
                                "DataType": "String",
                                "StringValue": self.sender_id,
                            }
                        },
                    )
                else:
                    client.publish(
                        TopicArn=self.topic, Subject=subject, Message=message
                    )
            except ClientError:
                self.alerter_logger.exception("couldn't send notification")
        else:
            if subject is None:
                target = self.number
            else:
                target = self.topic
            self.alerter_logger.info("dry_run: would send notifiction to %s:", target)
            if subject is not None:
                self.alerter_logger.info("    Subject: %s", subject)
            self.alerter_logger.info("    Message: %s", message)

    def _describe_action(self) -> str:
        if self.topic:
            return "posting to SNS topic {topic}".format(topic=self.topic)
        if self.number:
            return "SMSing {target} via SNS".format(target=self.number)
        return "not sending anything via SNS"



"""
Logging for SimpleMonitor.

Loggers process every monitor, every iteration, to record their state in some fashion.
"""

import logging
import time
from typing import Any, Dict, List, Optional, Union, cast

import arrow

from ..Monitors.monitor import Monitor
from ..util import format_datetime, get_config_option, subclass_dict_handler


class Logger:
    """Abstract class basis for loggers."""

    logger_type = "unknown"

    supports_batch = False
    doing_batch = False
    batch_data = None  # type: Optional[Dict[str, Any]]
    connected = True
    _global_info = None  # type: Optional[Dict[str, Any]]

    def __init__(self, config_options: Dict[str, Any]) -> None:
        self._config_options = config_options
        self.name = cast(str, self.get_config_option("_name", default="unnamed"))
        self.logger_logger = logging.getLogger("simplemonitor.logger-" + self.name)
        self._dependencies = cast(
            List[str],
            self.get_config_option("depend", required_type="[str]", default=[]),
        )
        # only log for Monitors with one of these groups
        self._groups = self.get_config_option(
            "groups", required_type="[str]", default=["default"]
        )
        if self.batch_data is None:
            self.batch_data = {}
        self.tz = cast(Optional[str], self.get_config_option("tz", default="UTC"))
        self.dateformat = cast(
            Optional[str],
            self.get_config_option(
                "dateformat",
                required_type="str",
                allowed_values=["timestamp", "iso8601"],
                default="timestamp",
            ),
        )
        self.enabled = cast(
            bool, self.get_config_option("enabled", required_type="bool", default=True)
        )

        if self._global_info is None:
            self._global_info = {}
        self.heartbeat = cast(
            bool,
            self.get_config_option("heartbeat", required_type="bool", default=False),
        )

    def __enter__(self) -> None:
        """Context manager entry."""
        self.start_batch()

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        """Context manager exit."""
        self.end_batch()

    def set_global_info(self, info: dict) -> None:
        """Receive global info about the SimpleMonitor state.

        Includes but not limited to refresh interval, known remote instances, etc"""
        self._global_info = info

    def get_config_option(
        self,
        key: str,
        *,
        default: Any = None,
        required: bool = False,
        required_type: str = "str",
        allowed_values: Any = None,
        allow_empty: bool = True,
        minimum: Optional[Union[int, float]] = None,
        maximum: Optional[Union[int, float]] = None,
    ) -> Any:
        """Get a config value.

        Throws the right flavour exception if something is wrong."""
        return get_config_option(
            self._config_options,
            key,
            default=default,
            required=required,
            required_type=required_type,
            allowed_values=allowed_values,
            allow_empty=allow_empty,
            minimum=minimum,
            maximum=maximum,
        )

    def hup(self) -> None:
        """Close and reopen our log file, if supported.

        This should be overridden where needed."""
        return  # pragma: no cover

    def save_result2(self, name: str, monitor: Monitor) -> None:
        """Record a result.

        Subclasses must override this with their implementation."""
        raise NotImplementedError

    def _get_datestring(self) -> str:
        """Format the current datetime according to the dateformat setting and timezone."""
        if self.dateformat == "iso8601":
            return format_datetime(arrow.now(), self.tz)
        return str(int(time.time()))

    @property
    def dependencies(self) -> list:
        """The dependencies of this Logger."""
        return self._dependencies

    @dependencies.setter
    def dependencies(self, dependency_list: List[str]) -> None:
        if not isinstance(dependency_list, list):
            raise TypeError("dependency_list must be a list")
        self._dependencies = dependency_list

    def check_dependencies(self, failed_list: List[str]) -> bool:
        """Compare a list of failed monitors to our dependencies, and mark
        the Logger as offline if one failed"""
        self.connected = True
        for dependency in failed_list:
            if dependency in self._dependencies:
                self.connected = False
        return self.connected

    @property
    def groups(self) -> List[str]:
        """The groups this Logger belongs to."""
        return self._groups

    def start_batch(self) -> None:
        """Prepare to process a batch of results"""
        if not self.supports_batch:
            return
        if self.doing_batch:
            self.logger_logger.error(
                "starting a batch while one was already in progress"
            )
        self.batch_data = {}
        self.doing_batch = True

    def end_batch(self) -> None:
        """End receiving a batch of results and process them"""
        if not self.supports_batch:
            return
        if not self.doing_batch:
            self.logger_logger.error("ending a batch when one wasn't in progress")
        self.process_batch()
        self.doing_batch = False

    def process_batch(self) -> None:
        """Process the batched data.
        This is blank for the base class."""
        return  # pragma: no cover

    def describe(self) -> str:
        """Explain what this logger does.
        We don't throw NotImplementedError here as it won't show up until something breaks,
        and we don't want to randomly die then."""
        return "(Logger did not write an auto-biography.)"  # pragma: no cover

    def __str__(self) -> str:
        return self.describe()


(register, get_class, all_types) = subclass_dict_handler(
    "simplemonitor.Loggers.logger", Logger, "logger_type"
)



# type: ignore
import os
import unittest

from simplemonitor.Loggers.file import HTMLLogger
from simplemonitor.Monitors.monitor import MonitorNull


class TestHTMLLogger(unittest.TestCase):
    def setUp(self):
        self._test_path = "test_html"
        try:
            os.mkdir(self._test_path)
        except FileExistsError:
            pass

    def test_html_logger(self):
        config_options = {
            "folder": "test_html",
            "copy_resources": True,
            "filename": "status.html",
        }
        test_logger = HTMLLogger(config_options)
        monitor = MonitorNull()
        monitor.run_test()
        test_logger.start_batch()
        test_logger.save_result2("test", monitor)
        test_logger.process_batch()
        test_logger.end_batch()

        self.assertTrue(
            os.path.exists(os.path.join("test_html", "status.html")),
            "status.html was not created",
        )
        self.assertTrue(
            os.path.exists(os.path.join("test_html", "main.bundle.js")),
            "main.bundle.js was not copied",
        )

    def tearDown(self):
        if os.path.isdir(self._test_path):
            for filename in os.listdir(self._test_path):
                try:
                    os.unlink(os.path.join(self._test_path, filename))
                except IOError:
                    pass
            try:
                os.rmdir(self._test_path)
            except IOError:
                pass



"""
UniFi monitoring for SimpleMonitor
"""

import re
from typing import Dict, NoReturn, Union, cast

from paramiko.client import RejectPolicy, SSHClient
from paramiko.ssh_exception import SSHException

from .monitor import Monitor, register


@register
class MonitorUnifiFailover(Monitor):
    """Monitor USG WAN Failover"""

    monitor_type = "unifi_failover"

    def __init__(self, name: str, config_options: dict) -> None:
        if "gap" not in config_options:
            config_options["gap"] = 300  # 5 mins
        super().__init__(name, config_options)
        self._router_address = cast(
            str, self.get_config_option("router_address", required=True)
        )
        self._username = cast(
            str, self.get_config_option("router_username", required=True)
        )
        self._password = self.get_config_option("router_password", required=False)
        self._ssh_key = self.get_config_option("ssh_key", required=False)
        if self._ssh_key is None and self._password is None:
            raise ValueError("must specify only one of router_password or ssh_key")
        self._check_interface = cast(
            str, self.get_config_option("check_interface", default="eth2")
        )

    def run_test(self) -> Union[NoReturn, bool]:
        try:
            with SSHClient() as client:
                client.set_missing_host_key_policy(RejectPolicy)
                client.load_system_host_keys()
                client.connect(
                    hostname=self._router_address,
                    username=self._username,
                    password=self._password,
                    key_filename=self._ssh_key,
                )
                _, stdout, _ = client.exec_command(  # nosec
                    "sudo /usr/sbin/ubnt-hal wlbGetStatus"
                )
                data = {}  # type: Dict[str, Dict[str, str]]
                data_block = {}  # type: Dict[str, str]
                interface = ""
                for _line in stdout.readlines():
                    line = _line.strip()
                    matches = re.match(r"interface +: (\w+)", line)
                    if matches:
                        if interface != "" and len(data_block) > 0:
                            data[interface] = data_block
                            data_block = {}
                        interface = matches.group(1)
                        data_block = {}
                        continue
                    matches = re.match(r"carrier +: (\w+)", line)
                    if matches:
                        data_block["carrier"] = matches.group(1)
                        continue
                    matches = re.match(r"status +: (\w+)", line)
                    if matches:
                        data_block["status"] = matches.group(1)
                        continue
                    matches = re.match(r"gateway +: (\w+)", line)
                    if matches:
                        data_block["gateway"] = matches.group(1)
                if interface != "" and len(data_block) > 0:
                    data[interface] = data_block
        except SSHException as error:
            self.monitor_logger.exception("Failed to ssh to USG")
            return self.record_fail("Failed to ssh to USG: {}".format(error))
        if self._check_interface not in data:
            self.monitor_logger.debug("processed data was %s", data)
            return self.record_fail(
                "Could not get status for interface {}".format(self._check_interface)
            )
        if data[self._check_interface]["carrier"] != "up":
            return self.record_fail(
                "Interface {} carrier is in status {} (wanted 'up')".format(
                    self._check_interface, data[self._check_interface]["carrier"]
                )
            )
        if data[self._check_interface]["status"] != "failover":
            return self.record_fail(
                "Interface {} is in status {} (wanted 'failover')".format(
                    self._check_interface, data[self._check_interface]["status"]
                )
            )
        if data[self._check_interface]["gateway"] == "unknown":
            return self.record_fail(
                "Interface {} has gateway {}".format(
                    self._check_interface, data[self._check_interface]["gateway"]
                )
            )
        return self.record_success(
            "Interface {} is {} with status {}".format(
                self._check_interface,
                data[self._check_interface]["carrier"],
                data[self._check_interface]["status"],
            )
        )

    def describe(self) -> str:
        return "Checking USG at {} has interface {} up and not failed over".format(
            self._router_address, self._check_interface
        )


@register
class MonitorUnifiFailoverWatchdog(Monitor):
    """Monitor UniFi WAN watchdog"""

    monitor_type = "unifi_watchdog"

    def __init__(self, name: str, config_options: dict) -> None:
        if "gap" not in config_options:
            config_options["gap"] = 300  # 5 mins
        super().__init__(name, config_options)
        self._router_address = cast(
            str, self.get_config_option("router_address", required=True)
        )
        self._username = cast(
            str, self.get_config_option("router_username", required=True)
        )
        self._password = self.get_config_option("router_password", required=False)
        self._ssh_key = self.get_config_option("ssh_key", required=False)
        if self._ssh_key is None and self._password is None:
            raise ValueError("must specify only one of router_password or ssh_key")
        self._primary_interface = cast(
            str, self.get_config_option("primary_interface", default="pppoe0")
        )
        self._secondary_interface = cast(
            str, self.get_config_option("secondary_interface", default="eth2")
        )

    def run_test(self) -> Union[NoReturn, bool]:
        try:
            with SSHClient() as client:
                client.set_missing_host_key_policy(RejectPolicy)
                client.load_system_host_keys()
                client.connect(
                    hostname=self._router_address,
                    username=self._username,
                    password=self._password,
                    key_filename=self._ssh_key,
                )
                _, stdout, _ = client.exec_command(  # nosec
                    "/usr/sbin/ubnt-hal wlbGetWdStatus"
                )
                data = {}  # type: Dict[str, Dict[str, str]]
                data_block = {}  # type: Dict[str, str]
                interface = ""
                for _line in stdout.readlines():
                    line = _line.strip()
                    matches = re.match(
                        r"([a-z]+[0-9])", line
                    )  # two spaces and an if name
                    if matches:
                        if interface != "" and len(data_block) > 0:
                            data[interface] = data_block
                            data_block = {}
                        interface = matches.group(1)
                        data_block = {}
                        continue
                    matches = re.match(r"status: (\w+)", line)
                    if matches:
                        data_block["status"] = matches.group(1)
                        continue
                    matches = re.match(r"ping gateway: ([^ ]+) - (\w+)", line)
                    if matches:
                        data_block["gateway"] = matches.group(1)
                        data_block["ping_status"] = matches.group(2)
                        continue
                if interface != "" and len(data_block) > 0:
                    data[interface] = data_block
        except SSHException as error:
            self.monitor_logger.exception("Failed to ssh to USG")
            return self.record_fail("Failed to ssh to USG: {}".format(error))
        for interface in [self._primary_interface, self._secondary_interface]:
            if interface not in data:
                self.monitor_logger.debug("processed data was %s", data)
                return self.record_fail(
                    "Could not get status for interface {}".format(interface)
                )
            if data[interface]["status"] != "Running":
                return self.record_fail(
                    "Interface {} in status {} (wanted 'Running')".format(
                        interface, data[interface]["status"]
                    )
                )
            if data[interface]["ping_status"] != "REACHABLE":
                return self.record_fail(
                    "Interface {} ping ({}) is {} (wanted 'REACHABLE')".format(
                        interface,
                        data[interface]["gateway"],
                        data[interface]["ping_status"],
                    )
                )
        return self.record_success(
            "Interfaces {} and {} both running and pinging".format(
                self._primary_interface, self._secondary_interface
            )
        )

    def describe(self) -> str:
        return "Checking USG at {} has interface {} and {} running and pinging".format(
            self._router_address, self._primary_interface, self._secondary_interface
        )



import unittest
from unittest.mock import MagicMock, patch

from simplemonitor.Alerters import HealthchecksAlerter


class TestHealthchecksAlerter(unittest.TestCase):
    @patch("requests.post")
    def test_send_notification(self, mock_post):
        """Test sending a notification successfully."""
        # Mock the response from requests
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_post.return_value = mock_response

        # Instantiate the Alerter
        config_options = {"token": "123456789"}

        alerter = HealthchecksAlerter(config_options)

        # Test the send_notification method
        alerter.send_notification(
            subject="Test Subject",
            body="Test Body",
            alert_type="success",
            name="Test Monitor",
            slug="test-slug",
        )

        # Check that requests.post was called with the correct URL
        expected_url = "https://hc-ping.com/123456789/test-slug"
        mock_post.assert_called_once_with(
            expected_url,
            data="Test Body",
            headers={"User-Agent": "SimpleMonitor"},
            timeout=5,
        )

    @patch("requests.post")
    def test_send_notification_create(self, mock_post):
        """Test sending a notification with the create option set to True"""
        # Mock the response from requests
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_post.return_value = mock_response

        config_options = {"token": "123456789"}
        config_options["create"] = True

        # Instantiate the Alerter
        alerter = HealthchecksAlerter(config_options)

        # Test the send_notification method
        alerter.send_notification(
            subject="Test Subject",
            body="Test Body",
            alert_type="success",
            name="Test Monitor",
            slug="test-slug",
        )

        # Check that requests.post was called with the correct URL
        expected_url = "https://hc-ping.com/123456789/test-slug?create=1"
        mock_post.assert_called_once_with(
            expected_url,
            data="Test Body",
            headers={"User-Agent": "SimpleMonitor"},
            timeout=5,
        )

    @patch("requests.post")
    def test_send_notification_failed(self, mock_post):
        """Test sending a notification successfully."""
        # Mock the response from requests
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_post.return_value = mock_response

        # Instantiate the Alerter
        config_options = {"token": "123456789"}

        alerter = HealthchecksAlerter(config_options)

        # Test the send_notification method
        alerter.send_notification(
            subject="Test Subject",
            body="Test Body",
            alert_type="failed",
            name="Test Monitor",
            slug="test-slug",
        )

        # Check that requests.post was called with the correct URL
        expected_url = "https://hc-ping.com/123456789/test-slug/fail"
        mock_post.assert_called_once_with(
            expected_url,
            data="Test Body",
            headers={"User-Agent": "SimpleMonitor"},
            timeout=5,
        )

    @patch("requests.post")
    def test_send_notification_failed_create(self, mock_post):
        """Test sending a notification successfully."""
        # Mock the response from requests
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_post.return_value = mock_response

        # Instantiate the Alerter
        config_options = {"token": "123456789"}
        config_options["create"] = True

        alerter = HealthchecksAlerter(config_options)

        # Test the send_notification method
        alerter.send_notification(
            subject="Test Subject",
            body="Test Body",
            alert_type="failed",
            name="Test Monitor",
            slug="test-slug",
        )

        # Check that requests.post was called with the correct URL
        expected_url = "https://hc-ping.com/123456789/test-slug/fail?create=1"
        mock_post.assert_called_once_with(
            expected_url,
            data="Test Body",
            headers={"User-Agent": "SimpleMonitor"},
            timeout=5,
        )


if __name__ == "__main__":
    unittest.main()



"""
Monitor things on a host for SimpleMonitor
"""

import json
import os
import re
import shlex
import subprocess  # nosec
import time
from typing import Tuple, cast

from markupsafe import escape

from ..util import bytes_to_size_string, size_string_to_bytes
from .monitor import Monitor, register

try:
    import psutil
except ImportError:
    psutil = None


try:
    import win32api

    WIN32_AVAILABLE = True
except ImportError:
    WIN32_AVAILABLE = False


@register
class MonitorDiskSpace(Monitor):
    """Make sure we have enough disk space."""

    monitor_type = "diskspace"

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        if self.is_windows(allow_cygwin=False):
            self.use_statvfs = False
            if not WIN32_AVAILABLE:
                raise RuntimeError(
                    "win32api is not available, but is needed for DiskSpace monitor."
                )
        else:
            self.use_statvfs = True
        self.partition = self.get_config_option("partition", required=True)
        self.limit = size_string_to_bytes(
            self.get_config_option("limit", required=True)
        )

    def run_test(self) -> bool:
        try:
            if self.use_statvfs:
                result = os.statvfs(self.partition)
                space = result.f_bavail * result.f_frsize
                percent = float(result.f_bavail) / float(result.f_blocks) * 100
            else:
                win_result = win32api.GetDiskFreeSpaceEx(self.partition)
                space = win_result[2]
                percent = float(win_result[2]) / float(win_result[1]) * 100
        except Exception as error:
            return self.record_fail("Couldn't get free disk space: %s" % error)

        if self.limit and space <= self.limit:
            return self.record_fail(
                "%s free (%d%%)" % (bytes_to_size_string(space), percent)
            )
        return self.record_success(
            "%s free (%d%%)" % (bytes_to_size_string(space), percent)
        )

    def describe(self) -> str:
        """Explains what we do."""
        if self.limit is None:
            limit = "none"
        else:
            limit = bytes_to_size_string(self.limit)

        return "Checking for at least %s free space on %s" % (limit, self.partition)

    def get_params(self) -> Tuple:
        return (self.limit, self.partition)


@register
class MonitorFileStat(Monitor):
    """Make sure a file exists, isn't too old and/or isn't too small."""

    monitor_type = "filestat"

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        self.maxage = self.get_config_option("maxage", required_type="int", minimum=0)
        _minsize = self.get_config_option(
            "minsize", required_type="str", allow_empty=True
        )
        if _minsize:
            self.minsize = size_string_to_bytes(_minsize)
        else:
            self.minsize = None
        _maxsize = self.get_config_option(
            "maxsize", required_type="str", allow_empty=True, default=""
        )
        if _maxsize:
            self.maxsize = size_string_to_bytes(_maxsize)
        else:
            self.maxsize = None
        self.filename = self.get_config_option("filename", required=True)

    def run_test(self) -> bool:
        try:
            statinfo = os.stat(self.filename)
        except FileNotFoundError:
            return self.record_fail("File %s does not exist" % self.filename)
        except Exception as error:
            return self.record_fail("Unable to check file: %s" % error)

        if self.minsize:
            if statinfo.st_size < self.minsize:
                return self.record_fail(
                    "Size is %d, should be >= %d bytes"
                    % (statinfo.st_size, self.minsize)
                )

        if self.maxsize:
            if statinfo.st_size > self.maxsize:
                return self.record_fail(
                    "Size is %d, should be <= %d bytes"
                    % (statinfo.st_size, self.maxsize)
                )

        now = time.time()
        diff = now - statinfo.st_mtime
        if self.maxage:
            if diff > self.maxage:
                return self.record_fail(
                    "Age is %d, should be < %d seconds" % (diff, self.maxage)
                )

        return self.record_success(
            "File {} exists (age: {}, size: {})".format(
                self.filename, int(diff), statinfo.st_size
            )
        )

    def describe(self) -> str:
        """Explains what we do"""
        desc = "Checking %s exists" % self.filename
        if self.maxage:
            desc = desc + " and is not older than %d seconds" % self.maxage
        if self.minsize:
            desc = desc + " and is not smaller than %d bytes" % self.minsize
        return desc

    def get_params(self) -> Tuple:
        return (self.filename, self.minsize, self.maxage)


@register
class MonitorApcupsd(Monitor):
    """Monitor an APC UPS (with apcupsd) to make sure it's ONLINE.

    Note: You must have apcupsd successfully setup and working for this monitor
    to function.
    """

    monitor_type = "apcupsd"

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        self.path = self.get_config_option("path", default="")

    def run_test(self) -> bool:
        info = {}
        if self.path != "":
            executable = os.path.join(self.path, "apcaccess")
        else:
            if self.is_windows():
                executable = os.path.join("c:\\", "apcupsd", "bin", "apcaccess.exe")
            else:
                executable = "apcaccess"
        try:
            _output = subprocess.check_output(executable)  # nosec
            output = _output.decode("utf-8")  # type: str
        except subprocess.CalledProcessError as error:
            output = error.output
        except OSError as error:
            return self.record_fail("Could not run {0}: {1}".format(executable, error))
        except Exception as error:
            return self.record_fail("Error while getting UPS info: {0}".format(error))

        for line in output.splitlines():
            if line.find(":") > -1:
                bits = line.split(":")
                info[bits[0].strip()] = bits[1].strip()

        if "STATUS" not in info:
            return self.record_fail("Could not get UPS status")

        if info["STATUS"] != "ONLINE":
            if "TIMELEFT" in info:
                return self.record_fail(
                    "%s: %s left" % (info["STATUS"], info["TIMELEFT"])
                )
            return self.record_fail(info["STATUS"])

        data = ""
        if "TIMELEFT" in info:
            data = "%s left" % (info["TIMELEFT"])

        if "LOADPCT" in info:
            if data != "":
                data += "; "
            data += "%s%% load" % info["LOADPCT"][0:4]

        return self.record_success(data)

    def describe(self) -> str:
        return "Monitoring UPS to make sure it's ONLINE."

    def get_params(self) -> Tuple:
        return (self.path,)


@register
class MonitorNUT(Monitor):
    """Check a NUT UPS"""

    monitor_type = "nut"

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        self.ups = cast(
            str, self.get_config_option("ups", required=True, allow_empty=False)
        )

    def run_test(self) -> bool:
        info = {}
        try:
            _output = subprocess.check_output(["upsc", self.ups])  # nosec
            output = _output.decode("utf-8")  # type: str
        except subprocess.CalledProcessError as error:
            output = error.output
        except OSError as error:
            return self.record_fail(f"Could not run upsc: {error}")
        except Exception as error:
            return self.record_fail(f"Error while getting UPS info: {error}")

        for line in output.splitlines():
            if line.find(":") > -1:
                bits = line.split(":")
                info[bits[0].strip()] = bits[1].strip()

        try:
            ups_status = cast(str, info["ups.status"])
        except KeyError:
            return self.record_fail("Could not get UPS status")

        status_tokens = ups_status.split()
        ok = True
        message: list[str] = []
        if "OB" in status_tokens:
            message.append("UPS is on battery")
            ok = False
        if "OL" in status_tokens:
            message.append("online")
        if "RB" in status_tokens:
            message.append("battery needs replacing!")
            ok = False
        if "ups.load" in info:
            message.append(f"load: {info['ups.load']}%")
        if "battery.charge" in info:
            message.append(f"charge: {info['battery.charge']}")
        if "battery.runtime" in info:
            message.append(f"runtime: {info['battery.runtime']}")

        message_str = "; ".join(message)
        if ok:
            return self.record_success(message_str)
        return self.record_fail(message_str)

    def describe(self) -> str:
        return f"Monitoring UPS {self.ups} to make sure it's online"

    def get_params(self) -> Tuple:
        return (self.ups,)


@register
class MonitorPortAudit(Monitor):
    """Check a host doesn't have outstanding security issues."""

    monitor_type = "portaudit"
    regexp = re.compile(r"(\d+) problem\(s\) in your installed packages found")

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        self.path = self.get_config_option("path", default="")

    def describe(self) -> str:
        return "Checking for insecure ports."

    def get_params(self) -> Tuple:
        return (self.path,)

    def run_test(self) -> bool:
        try:
            # -X 1 tells portaudit to re-download db if one day out of date
            if self.path == "":
                self.path = "/usr/local/sbin/portaudit"
            try:
                # nosec
                _output = subprocess.check_output([self.path, "-a", "-X", "1"])  # nosec
                output = _output.decode("utf-8")
            except subprocess.CalledProcessError as error:
                output = error.output
            except OSError as error:
                return self.record_fail("Error running %s: %s" % (self.path, error))
            except Exception as error:
                return self.record_fail("Error running portaudit: %s" % error)

            for line in output.splitlines():
                matches = self.regexp.match(line)
                if matches:
                    count = int(matches.group(1))
                    # sanity check
                    if count == 0:
                        return self.record_success()
                    if count == 1:
                        return self.record_fail("1 problem")
                    return self.record_fail("%d problems" % count)
            return self.record_success()
        except Exception as error:
            return self.record_fail("Could not run portaudit: %s" % error)


@register
class MonitorPkgAudit(Monitor):
    """Check a host doesn't have outstanding security issues."""

    monitor_type = "pkgaudit"
    regexp = re.compile(r"(\d+) problem\(s\) in \w+ installed package(s|\(s\)) found")
    path = ""

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        self.path = self.get_config_option("path", default="")

    def describe(self) -> str:
        return "Checking for insecure packages."

    def get_params(self) -> Tuple:
        return (self.path,)

    def run_test(self) -> bool:
        if self.path == "":
            self.path = "/usr/local/sbin/pkg"
        try:
            _output = subprocess.run(
                [self.path, "audit", "--raw=json"], capture_output=True
            )  # nosec
            output = json.loads(_output.stdout.decode("utf-8"))
        except json.JSONDecodeError as error:
            return self.record_fail(f"Failed to decode JSON output: {error}")
        except OSError as error:
            return self.record_fail(f"Failed to run {self.path} audit: {error}")
        except Exception as error:
            return self.record_fail(f"Error running pkg audit: {error}")

        try:
            count = int(output["pkg_count"])
        except KeyError:
            return self.record_fail("Failed to find pkg_count in output")
        if count == 0:
            return self.record_success()
        if count == 1:
            return self.record_fail("1 problem found")
        return self.record_fail("%d problems found" % count)


@register
class MonitorLoadAvg(Monitor):
    """Check a host's load average isn't too high."""

    monitor_type = "loadavg"

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        if self.is_windows(allow_cygwin=False):
            raise RuntimeError("loadavg monitor does not support Windows")
        # which time field we're looking at: 0 = 1min, 1 = 5min, 2=15min
        self.which = self.get_config_option(
            "which", required_type="int", default=1, minimum=0, maximum=2
        )
        self.max = self.get_config_option(
            "max", required_type="float", default=1.00, minimum=0
        )

    def describe(self) -> str:
        if self.which == 0:
            return "Checking 1min loadavg is <= %0.2f" % self.max
        if self.which == 1:
            return "Checking 5min loadavg is <= %0.2f" % self.max
        return "Checking 15min loadavg is <= %0.2f" % self.max

    def run_test(self) -> bool:
        try:
            loadavg = os.getloadavg()
        except Exception as error:
            return self.record_fail("Exception getting loadavg: %s" % error)

        if loadavg[self.which] > self.max:
            return self.record_fail("%0.2f" % loadavg[self.which])
        return self.record_success("%0.2f" % loadavg[self.which])

    def get_params(self) -> Tuple:
        return (self.which, self.max)


@register
class MonitorMemory(Monitor):
    """Check for available memory."""

    monitor_type = "memory"

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        if psutil is None:
            self.monitor_logger.critical("psutil is not installed.")
            self.monitor_logger.critical("Try: pip install -r requirements.txt")
        self.percent_free = cast(
            int,
            self.get_config_option("percent_free", required_type="int", required=True),
        )

    def run_test(self) -> bool:
        if psutil is None:
            return self.record_fail("psutil is not installed")
        stats = psutil.virtual_memory()
        percent = int(stats.available / stats.total * 100)
        message = "{}% free".format(percent)
        if percent < self.percent_free:
            return self.record_fail(message)
        return self.record_success(message)

    def get_params(self) -> Tuple:
        return (self.percent_free,)

    def describe(self) -> str:
        return "Checking for at least {}% free memory".format(self.percent_free)


@register
class MonitorSwap(Monitor):
    """Check for available swap."""

    monitor_type = "swap"

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        if psutil is None:
            self.monitor_logger.critical("psutil is not installed.")
            self.monitor_logger.critical("Try: pip install -r requirements.txt")
        self.percent_free = cast(
            int,
            self.get_config_option("percent_free", required_type="int", required=True),
        )

    def run_test(self) -> bool:
        if psutil is None:
            return self.record_fail("psutil is not installed")
        stats = psutil.swap_memory()
        percent = 100 - stats.percent
        message = f"{percent:.2f}% free"
        if percent < self.percent_free:
            return self.record_fail(message)
        return self.record_success(message)

    def get_params(self) -> Tuple:
        return (self.percent_free,)

    def describe(self) -> str:
        return f"Checking for at least {self.percent_free}% free swap"


@register
class MonitorZap(Monitor):
    """Checks a Zap channel to make sure it is ok"""

    monitor_type = "zap"
    r = re.compile("^alarms=(?P<status>).+")

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        self.span = self.get_config_option(
            "span", required_type="int", default=1, minimum=1
        )

    def run_test(self) -> bool:
        try:
            _output = subprocess.check_output(["ztscan", str(self.span)])  # nosec
            output = _output.decode("utf-8")
            for line in output:
                matches = self.r.match(line)
                if matches:
                    status = matches.group("status")
                    if status != "OK":
                        return self.record_fail("status is %s" % status)
                    return self.record_success()
            return self.record_fail("Error getting status")
        except Exception as error:
            return self.record_fail("Error running ztscan: %s" % error)

    def describe(self) -> str:
        return "Checking status of zap span %d is OK" % self.span

    def get_params(self) -> Tuple:
        return (self.span,)


@register
class MonitorCommand(Monitor):
    """Check the output of a command.

    We can check for a regexp match in the output or give a max value and check the
    output is lower than this value.
    """

    result_regexp = None
    monitor_type = "command"

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)

        self.result_regexp_text = self.get_config_option("result_regexp", default="")
        self.result_max = self.get_config_option("result_max", required_type="int")
        if self.result_regexp_text != "":
            self.result_regexp = re.compile(self.result_regexp_text)
            if self.result_max is not None:
                self.monitor_logger.error(
                    "command monitors do not support result_regexp AND"
                    "result_max settings simultaneously"
                )
                self.result_max = None
        self.show_output = self.get_config_option(
            "show_output", required_type="bool", default=False
        )

        command = self.get_config_option("command", required=True, allow_empty=False)
        self.command = shlex.split(command)

    def run_test(self) -> bool:
        try:
            _out = subprocess.check_output(self.command)  # nosec
            if self.result_regexp is not None:
                out = _out.decode("utf-8")
                matches = self.result_regexp.search(out)
                if matches:
                    return self.record_success()
                return self.record_fail("could not match regexp in out")
            if self.result_max is not None:
                outasinteger = int(_out)
                if outasinteger < self.result_max:
                    return self.record_success(
                        "%s < %s" % (outasinteger, self.result_max)
                    )
                return self.record_fail("%s >= %s" % (outasinteger, self.result_max))
            msg = ""
            if self.show_output:
                msg = escape(_out.decode("utf-8"))
            return self.record_success(msg)
        except subprocess.CalledProcessError as exception:
            if self.show_output:
                return self.record_fail(exception.output.decode("utf-8"))
            return self.record_fail(str(exception))
        except Exception as exception:
            return self.record_fail(str(exception))

    def describe(self) -> str:
        """Explains what this instance is checking"""
        if self.result_regexp is not None:
            return 'checking command "%s" match a regexp %s' % (
                " ".join(self.command),
                self.result_regexp_text,
            )
        if self.result_max is not None:
            return 'checking command "%s" returns a value < %d' % (
                " ".join(self.command),
                self.result_max,
            )
        return "checking command '%s' has return status 0" % " ".join(self.command)

    def get_params(self) -> Tuple:
        return (
            self.command,
            self.result_regexp_text,
            self.result_max,
            self.show_output,
        )



"""
SimpleMonitor alerts via Telegram
"""

from typing import cast

import requests

from ..Monitors.monitor import Monitor
from .alerter import Alerter, AlertLength, AlertType, register


@register
class TelegramAlerter(Alerter):
    """Send push notification via Telegram."""

    alerter_type = "telegram"

    def __init__(self, config_options: dict) -> None:
        super().__init__(config_options)

        self.telegram_token = cast(
            str, self.get_config_option("token", required=True, allow_empty=False)
        )

        self.telegram_chatid = cast(
            str, self.get_config_option("chat_id", required=True, allow_empty=False)
        )

        self.timeout = cast(
            int, self.get_config_option("timeout", required_type="int", default=5)
        )

        self.support_catchup = True

    def send_telegram_notification(self, body: str) -> None:
        """Send a push notification."""

        response = requests.post(
            f"https://api.telegram.org/bot{self.telegram_token}/sendMessage",
            data={"chat_id": self.telegram_chatid, "text": body},
            timeout=self.timeout,
        )
        if not response.status_code == requests.codes.ok:
            raise RuntimeError("Unable to send telegram notification")

    def send_alert(self, name: str, monitor: Monitor) -> None:
        """Build up the content for the push notification."""

        alert_type = self.should_alert(monitor)
        if alert_type == AlertType.NONE:
            return

        body = self.build_message(AlertLength.FULL, alert_type, monitor)

        if not self._dry_run:
            try:
                self.send_telegram_notification(body)
                self.alerter_logger.info(
                    f"monitor: {name} | alerter: {self.name} - sent"
                )
            except Exception as e:
                self.alerter_logger.exception(f"Couldn't send push notification: {e}")
        else:
            self.alerter_logger.info("dry_run: would send push notification: %s", body)

    def _describe_action(self) -> str:
        return f"posting messages to {self.telegram_chatid} on Telegram"



"""
Network logging support for SimpleMonitor
"""

import datetime
import hmac
import logging
import socket
import struct
from threading import Thread
from typing import TYPE_CHECKING, Dict, Optional, TypedDict, Union, cast

from ..Monitors.monitor import Monitor
from ..util import LoggerConfigurationError
from ..util.json_encoding import json_dumps, json_loads
from .logger import Logger, register

if TYPE_CHECKING:
    from ..simplemonitor import SimpleMonitor

# From the docs:
#  Threads interact strangely with interrupts: the KeyboardInterrupt exception
#  will be received by an arbitrary thread. (When the signal module is
#  available, interrupts always go to the main thread.)

_DIGEST_NAME = "md5"


class RemoteHost(TypedDict):
    last_seen: datetime.datetime
    address: str


@register
class NetworkLogger(Logger):
    """Send our results over the network to another instance."""

    logger_type = "network"
    supports_batch = True

    def __init__(self, config_options: dict) -> None:
        super().__init__(config_options)

        self.host = cast(
            str, self.get_config_option("host", required=True, allow_empty=False)
        )
        self.port = cast(
            int, self.get_config_option("port", required_type="int", required=True)
        )
        self.hostname = cast(str, self.get_config_option("client_name"))
        self.key = bytearray(
            self.get_config_option("key", required=True, allow_empty=False), "utf-8"
        )

    def describe(self) -> str:
        return "Sending monitor results to {0}:{1}".format(self.host, self.port)

    def save_result2(self, name: str, monitor: Monitor) -> None:
        if not self.doing_batch:  # pragma: no cover
            self.logger_logger.error(
                "NetworkLogger.save_result2() called while not doing batch."
            )
            return
        self.logger_logger.debug("network logger: %s %s", name, monitor)
        if monitor.monitor_type == "unknown":
            self.logger_logger.error(
                "Cannot serialize monitor %s, has type 'unknown'.", name
            )
            return
        try:
            if monitor.monitor_type == "compound":
                self.logger_logger.error(
                    "not pickling compound monitor - currently incompatible with network loggers"
                )
            else:
                data = {
                    "cls_type": monitor.monitor_type,
                    "data": monitor.to_python_dict(),
                }
                if self.batch_data is not None:
                    self.batch_data[monitor.name] = data
                else:
                    self.batch_data = {monitor.name: data}
        except Exception:  # pylint: disable=broad-except
            self.logger_logger.exception("Failed to serialize monitor %s", name)

    def process_batch(self) -> None:
        try:
            payload = json_dumps(
                {
                    "version": 2,
                    "name": self.hostname,
                    "monitors": self.batch_data,
                }
            )
            mac = hmac.new(self.key, payload, _DIGEST_NAME)
            send_bytes = struct.pack("B", mac.digest_size) + mac.digest() + payload
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                try:
                    sock.connect((self.host, self.port))
                except socket.error:
                    sock.close()
                    sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
                    sock.connect((self.host, self.port))
                sock.send(send_bytes)
            finally:
                sock.close()
        except Exception as exception:  # pylint: disable=broad-except
            self.logger_logger.exception("Failed to send network data: %s", exception)


class Listener(Thread):
    """
    Handle incoming remote connections.

    This class isn't actually a Logger, but is the receiving-end
    implementation for network logging.

    Here seemed a reasonable place to put it."""

    def __init__(
        self,
        simplemonitor: "SimpleMonitor",
        port: int,
        key: Optional[str] = None,
        bind_host: str = "",
        ipv4_only: bool = False,
    ) -> None:
        """Set up the thread.

        simplemonitor is a SimpleMonitor object which we will put our results into.
        """
        if key is None or key == "":
            raise LoggerConfigurationError("Network logger key is missing")
        Thread.__init__(self, daemon=True)
        if ipv4_only:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        else:
            # try IPv6 and fallback to IPv4
            try:
                self.sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
                self.sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, False)
            except OSError:
                self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind((bind_host, port))
        self.simplemonitor = simplemonitor
        self.key = bytearray(key, "utf-8")
        self.logger = logging.getLogger("simplemonitor.logger.networklistener")
        self.running = False  # type: bool

    def run(self) -> None:
        """The main body of our thread.

        The loop here keeps going until we're killed by the main app.
        When the main app kills us (with join()), socket.listen throws socket.error.
        """
        self.running = True
        conn: Optional[socket.socket] = None
        addr: Optional[socket._RetAddress] = None
        while self.running:
            try:
                self.sock.listen(5)
                conn, addr = self.sock.accept()
                conn.settimeout(5.0)
                assert conn is not None
                assert addr is not None
                self.logger.debug("Got connection from %s", addr[0])
                serialized = bytearray()
                while 1:
                    data = conn.recv(1024)
                    if not data:
                        break
                    serialized += data
                conn.close()
                if len(serialized) == 0:
                    self.logger.debug("No data from %s", addr[0])
                    continue
                self.logger.debug("Finished receiving from %s", addr[0])
                try:
                    # first byte is the size of the MAC
                    mac_size = serialized[0]
                    # then the MAC
                    their_digest = serialized[1 : mac_size + 1]
                    # then the rest is the serialized data
                    serialized = serialized[mac_size + 1 :]
                    mac = hmac.new(self.key, serialized, _DIGEST_NAME)
                    my_digest = mac.digest()
                except IndexError as error:  # pragma: no cover
                    raise ValueError(
                        "Did not receive any or enough data from {}".format(addr[0])
                    ) from error
                if isinstance(my_digest, str):
                    self.logger.debug(
                        "Computed my digest to be %s; remote is %s",
                        my_digest,
                        their_digest,
                    )
                else:
                    self.logger.debug(
                        "Computed my digest to be %s; remote is %s",
                        my_digest.hex(),
                        their_digest.hex(),
                    )
                if not hmac.compare_digest(their_digest, my_digest):
                    raise Exception(
                        "Mismatched MAC for network logging data from %s\n"
                        "Mismatched key? Old version of SimpleMonitor?\n" % addr[0]
                    )
                result = json_loads(bytes(serialized))  # type: dict
                version = result.get("version", 1)
                if version == 1:
                    self.logger.debug("Received version 1 data from %s", addr[0])
                    self.simplemonitor.update_remote_monitor(result, addr[0])
                elif version == 2:
                    self.logger.debug("Received version 2 data from %s", addr[0])
                    self._handle_data_v2(result, addr[0])
                else:
                    self.logger.critical(
                        "Received unknown version %s data from %s cannot process",
                        version,
                        addr[0],
                    )
            except socket.timeout:
                if addr:
                    self.logger.warning("Timeout during recv from %s", addr[0])
                if conn:
                    conn.close()
            except socket.error as exception:
                if exception.errno == 4:
                    # Interrupted system call
                    self.logger.warning(
                        "Interrupted system call in thread, I think that's a ^C"
                    )
                    self.running = False
                    self.sock.close()
                if self.running:
                    self.logger.exception("Socket error caught in thread")
            except Exception:  # pylint: disable=broad-except
                self.logger.exception("Listener thread caught exception")
        self.logger.warning("Listener stopped")

    def _handle_data_v2(
        self, data: Dict[str, Union[str, int, Dict[str, dict]]], source: str
    ) -> None:
        """Handle data in v2 format

        {
            "version": 2,
            "name": "remote_instance_name",
            "monitors": [ monitor data, ... ]
        }
        """
        remote_instance_name = str(data.get("name", source))
        if not remote_instance_name or remote_instance_name == "None":
            remote_instance_name = source
        remote_monitors = data.get("monitors", None)
        if remote_monitors is None:
            self.logger.error(
                "Received empty monitors list from remote instance %s",
                remote_instance_name,
            )
        elif isinstance(remote_monitors, dict):
            self.simplemonitor.update_remote_monitor(
                remote_monitors, remote_instance_name
            )
            self.simplemonitor.upsert_remote_host(
                remote_instance_name, datetime.datetime.now(), source
            )
        else:
            self.logger.error(
                "Bad data type for monitors from remote instance %s",
                remote_instance_name,
            )



"""
SimpleMonitor alerts via macOS Notification Center
"""

import platform

try:
    import pync

    PYNC_AVAILABLE = True
except ImportError:
    PYNC_AVAILABLE = False

from ..Monitors.monitor import Monitor
from .alerter import Alerter, AlertLength, AlertType, register


@register
class NotificationCenterAlerter(Alerter):
    """Send alerts to the macOS Notification Center"""

    alerter_type = "nc"

    def __init__(self, config_options: dict) -> None:
        super().__init__(config_options)
        if platform.system() != "Darwin" or not PYNC_AVAILABLE:
            self.alerter_logger.critical(
                "This alerter (currently) only works on Mac OS X!"
            )
            return

    def send_alert(self, name: str, monitor: Monitor) -> None:
        """Send the message"""
        if not PYNC_AVAILABLE:
            self.alerter_logger.critical("Missing pync package")
            return
        alert_type = self.should_alert(monitor)
        message = ""

        if alert_type not in [AlertType.FAILURE, AlertType.CATCHUP]:
            return
        message = self.build_message(AlertLength.NOTIFICATION, alert_type, monitor)

        if not self._dry_run:
            pync.notify(message=message, title="SimpleMonitor")
        else:
            self.alerter_logger.info("dry_run: would send message: %s", message)

    def _describe_action(self) -> str:
        return "sending notifications via Notification Center"



# coding=utf-8
"""Execution logic for SimpleMonitor."""

import concurrent.futures
import copy
import logging
import os
import signal
import sys
import time
from datetime import datetime
from pathlib import Path
from socket import gethostname
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union, cast

from .Alerters.alerter import Alerter
from .Alerters.alerter import all_types as all_alerter_types
from .Alerters.alerter import get_class as get_alerter_class
from .Loggers.logger import Logger
from .Loggers.logger import all_types as all_logger_types
from .Loggers.logger import get_class as get_logger_class
from .Loggers.network import Listener, RemoteHost
from .Monitors.compound import CompoundMonitor
from .Monitors.monitor import Monitor, MonitorState
from .Monitors.monitor import all_types as all_monitor_types
from .Monitors.monitor import get_class as get_monitor_class
from .util import check_group_match, get_config_dict
from .util.envconfig import EnvironmentAwareConfigParser

module_logger = logging.getLogger("simplemonitor")


class SimpleMonitor:
    """A fairly simple monitor."""

    def __init__(
        self,
        config_file: Union[str, Path],
        *,
        hup_file: Optional[Path] = None,
        no_network: bool = False,
        max_loops: int = -1,
        heartbeat: bool = True,
        one_shot: bool = False,
        max_workers: Optional[int] = None,
    ) -> None:
        """Main class turn on."""
        if isinstance(config_file, str):
            self._config_file = Path(config_file)
        elif isinstance(config_file, Path):
            self._config_file = config_file
        else:
            raise ValueError("config_file must be str or Path")

        self.monitors = {}  # type: Dict[str, Monitor]
        self.failed = []  # type: List[str]
        self.still_failing = []  # type: List[str]
        self.skipped = []  # type: List[str]
        self.warning = []  # type: List[str]
        self.remote_monitors = {}  # type: Dict[str, Dict[str, Monitor]]

        self.loggers = {}  # type: Dict[str, Logger]
        self.alerters = {}  # type: Dict[str, Alerter]

        self._hup_file = hup_file
        self._need_hup = False
        self._hup_timestamp = None  # type: Optional[float]
        self._no_network = no_network
        self._remote_listening_thread = None  # type: Optional[Listener]
        self._max_loops = max_loops
        self.config_ok = False
        self.heartbeat = heartbeat
        self.one_shot = one_shot
        self.pidfile = None  # type: Optional[str]
        self._max_workers = max_workers
        self._remote_hosts: dict[str, RemoteHost] = {}

        self._setup_signals()
        self._load_config()

    def _load_config(self) -> None:
        """Load config, monitors, alerters and loggers."""

        config_ok = True
        config = EnvironmentAwareConfigParser()
        if not self._config_file.exists():
            raise RuntimeError(
                "Configuration file {} does not exist".format(self._config_file)
            )
        config.read(self._config_file)

        self.interval = config.getint("monitor", "interval")
        self.pidfile = config.get("monitor", "pidfile", fallback=None)
        hup_file = config.get("monitor", "hup_file", fallback=None)
        if hup_file is not None:
            self._hup_file = Path(hup_file)
            module_logger.info(
                "Watching modification time of %s; increase it to trigger a config reload",
                hup_file,
            )
            self._check_hup_file()

        if not self._no_network and config.get(
            "monitor", "remote", fallback="0"
        ).lower() in ["1", "true", "yes"]:
            self._network = True
            self._remote_port = int(config.get("monitor", "remote_port"))
            self._network_key = config.get("monitor", "key", fallback=None)
            self._network_bind_host = config.get("monitor", "bind_host", fallback="")
            self._ipv4_only = cast(
                bool, config.get("monitor", "ipv4_only", fallback=False)
            )
        else:
            self._network = False

        monitors_files = [
            Path(config.get("monitor", "monitors", fallback="monitors.ini"))
        ]
        monitors_dir = config.get("monitor", "monitors_dir", fallback=None)
        if monitors_files[0] == Path("."):
            module_logger.debug("No main monitors.ini file specified")
            monitors_files.pop(0)
        elif not monitors_files[0].exists():
            raise RuntimeError(
                f"Monitors configuration file '{monitors_files[0]}' does not exist"
            )
        if monitors_dir:
            monitors_files.extend(list(sorted(Path(monitors_dir).glob("*.ini"))))
        config_ok = self._load_monitors(monitors_files)

        count = self.count_monitors()
        if count == 0:
            module_logger.critical("No monitors loaded")
        config_ok = config_ok & self._load_loggers(config)
        config_ok = config_ok & self._load_alerters(config)
        if not self._verify_dependencies():
            raise RuntimeError("Broken dependency configuration")
        if not self.verify_alerting():
            module_logger.warning("No alerters defined and no remote logger found")
        self.config_ok = config_ok

    def _start_network_thread(self) -> None:
        if self._remote_listening_thread:
            # if the thread is running, check if it should be
            if not self._network:
                module_logger.info("Stopping remote listener thread")
                self._remote_listening_thread.running = False
            return
        if self._network:
            module_logger.info("Starting remote listener thread")
            self._remote_listening_thread = Listener(
                self,
                self._remote_port,
                self._network_key,
                bind_host=self._network_bind_host,
                ipv4_only=self._ipv4_only,
            )
            self._remote_listening_thread.start()

    def _load_monitors(self, filenames: Sequence[Union[Path, str]]) -> bool:
        """Load all the monitors from the config file."""

        all_good = True
        config = EnvironmentAwareConfigParser()
        config.read(filenames)
        module_logger.info(
            "Loaded monitor config from: %s",
            ", ".join(map(str, filenames)),
        )
        monitors = config.sections()
        if "defaults" in monitors:
            default_config = get_config_dict(config, "defaults")
            monitors.remove("defaults")
        else:
            default_config = {}

        myhostname = gethostname().lower()

        module_logger.info("=== Loading monitors")
        for this_monitor in monitors:
            if config.has_option(this_monitor, "runon"):
                if myhostname != config.get(this_monitor, "runon").lower():
                    module_logger.warning(
                        "Ignoring monitor %s because it's only for host %s",
                        this_monitor,
                        config.get(this_monitor, "runon"),
                    )
                    continue

            monitor_type = config.get(this_monitor, "type")

            new_monitor = None
            config_options = default_config.copy()
            config_options.update(get_config_dict(config, this_monitor))
            if self.has_monitor(this_monitor):
                if self.monitors[this_monitor].monitor_type == config_options["type"]:
                    module_logger.info(
                        "Updating configuration for monitor %s", this_monitor
                    )
                    self.update_monitor_config(this_monitor, config_options)
                else:
                    module_logger.error(
                        "Cannot update monitor %s from type %s to type %s. "
                        "Keeping original config for this monitor.",
                        this_monitor,
                        self.monitors[this_monitor].monitor_type,
                        config_options["type"],
                    )
                    all_good = False
                continue

            try:
                cls = get_monitor_class(monitor_type)
            except KeyError:
                module_logger.error(
                    "Unknown monitor type %s; valid types are: %s",
                    monitor_type,
                    ", ".join(all_monitor_types()),
                )
                all_good = False
                continue
            new_monitor = cls(this_monitor, config_options)
            # new_monitor.set_mon_refs(m)

            if new_monitor.enabled:
                module_logger.info(
                    "Adding %s monitor %s: %s", monitor_type, this_monitor, new_monitor
                )
                self.add_monitor(this_monitor, new_monitor)
            else:
                module_logger.info(
                    "Skipping disabled %s monitor: %s", monitor_type, this_monitor
                )
            # del new_monitor  # ?

        for monitor in self.monitors.values():
            monitor.set_mon_refs(self.monitors)
            monitor.set_sm_ref(self)
            monitor.post_config_setup()
        self.prune_monitors(monitors)
        if not all_good:
            module_logger.warning(
                "--- Loaded %d monitors but with errors", self.count_monitors()
            )
            return False
        module_logger.info("--- Loaded %d monitors", self.count_monitors())
        return True

    def _load_loggers(self, config: EnvironmentAwareConfigParser) -> bool:
        """Load the loggers listed in the config object."""

        all_good = True
        if config.has_option("reporting", "loggers"):
            loggers = config.get("reporting", "loggers").split(",")
        else:
            loggers = []

        module_logger.info("=== Loading loggers")
        for config_logger in loggers:
            logger_type = config.get(config_logger, "type")
            config_options = get_config_dict(config, config_logger)
            config_options["_name"] = config_logger
            if self.has_logger(config_logger):
                if self.loggers[config_logger].logger_type == config_options["type"]:
                    module_logger.info(
                        "Updating configuration for logger %s", config_logger
                    )
                    self.update_logger_config(config_logger, config_options)
                else:
                    module_logger.error(
                        "Cannot update logger %s from type %s to type %s. "
                        "Keeping original config for this logger.",
                        config_logger,
                        self.loggers[config_logger].logger_type,
                        config_options["type"],
                    )
                    all_good = False
                continue
            try:
                logger_cls = get_logger_class(logger_type)
            except KeyError:
                module_logger.error(
                    "Unknown logger type %s; valid types are: %s",
                    logger_type,
                    ", ".join(all_logger_types()),
                )
                all_good = False
                continue
            new_logger = logger_cls(config_options)  # type: Logger
            new_logger.set_global_info(
                {"interval": config.getint("monitor", "interval")}
            )
            if new_logger.enabled:
                module_logger.info(
                    "Adding %s logger %s: %s",
                    logger_type,
                    config_logger,
                    new_logger.describe(),
                )
                self.add_logger(config_logger, new_logger)
            else:
                module_logger.info(
                    "Skipping disabled %s logger: %s", logger_type, config_logger
                )

            del new_logger
        self.prune_loggers(loggers)
        if not all_good:
            module_logger.warning(
                "--- Loaded %d loggers but with errors", len(self.loggers)
            )
            return False
        module_logger.info("--- Loaded %d loggers", len(self.loggers))
        return True

    def _load_alerters(self, config: EnvironmentAwareConfigParser) -> bool:
        """Load the alerters listed in the config object."""
        all_good = True
        if config.has_option("reporting", "alerters"):
            alerters = config.get("reporting", "alerters").split(",")
        else:
            alerters = []

        module_logger.info("=== Loading alerters")
        for this_alerter in alerters:
            alerter_type = config.get(this_alerter, "type")
            config_options = get_config_dict(config, this_alerter)
            if self.has_alerter(this_alerter):
                if self.alerters[this_alerter].alerter_type == config_options["type"]:
                    module_logger.info(
                        "Updating configuration for alerter %s", this_alerter
                    )
                    self.update_alerter_config(this_alerter, config_options)
                else:
                    module_logger.error(
                        "Cannot update alerter %s from type %s to type %s. "
                        "Keeping original config for this alerter.",
                        this_alerter,
                        self.alerters[this_alerter].alerter_type,
                        config_options["type"],
                    )
                    all_good = False
                continue
            try:
                alerter_cls = get_alerter_class(alerter_type)
            except KeyError:
                module_logger.error(
                    "Unknown alerter type %s; valid types are: %s",
                    alerter_type,
                    ", ".join(all_alerter_types()),
                )
                all_good = False
                continue
            new_alerter = alerter_cls(config_options)  # type: Alerter

            if new_alerter.enabled:
                module_logger.info(
                    "Adding %s alerter %s: %s",
                    alerter_type,
                    this_alerter,
                    new_alerter.describe(),
                )
                new_alerter.name = this_alerter
                self.add_alerter(this_alerter, new_alerter)
            else:
                module_logger.info(
                    "Skipping disabled %s alerter: %s", alerter_type, this_alerter
                )

            del new_alerter
        self.prune_alerters(alerters)
        if not all_good:
            module_logger.info(
                "--- Loaded %d alerters but with errors", len(self.alerters)
            )
            return False
        module_logger.info("--- Loaded %d alerters", len(self.alerters))
        return True

    def _setup_signals(self) -> None:
        """Set up the SIGHUP handler."""
        _message = (
            "Unable to trap SIGHUP... maybe it doesn't exist on this platform. "
            "Set 'hup_file' in config and touch that file to trigger a config reload."
        )
        try:
            signal.signal(signal.SIGHUP, self._handle_sighup)
        except ValueError:  # pragma: no cover
            module_logger.warning(_message)
        except AttributeError:  # pragma: no cover
            module_logger.warning(_message)

    def _handle_sighup(self, *_: Any) -> None:
        """Receive SIGHUP and process it."""
        module_logger.warning("Received SIGHUP")
        self._need_hup = True

    def _check_hup_file(self) -> bool:
        """Check a file's timestamp, and if it's newer than last time, treat it
        the same as receiving SIGHUP so that a reload is triggered. This allows
        config reloading on platforms which don't support the signal (i.e.
        Windows)"""
        if self._hup_file is None:
            return False
        try:
            statinfo = os.stat(self._hup_file)
        except IOError:
            module_logger.debug(
                "Could not call stat() on path %s for file-based HUP", self._hup_file
            )
            return False
        modification_time = statinfo.st_mtime
        if self._hup_timestamp is None:
            self._hup_timestamp = modification_time
            return True
        if modification_time > self._hup_timestamp:
            self._hup_timestamp = modification_time
            return True
        return False

    def _create_pid_file(self) -> None:
        if self.pidfile:
            my_pid = os.getpid()
            try:
                with open(self.pidfile, "w") as file_handle:
                    file_handle.write("%d\n" % my_pid)
            except IOError:
                module_logger.error("Couldn't write to pidfile!")
                self.pidfile = None

    def _remove_pid_file(self) -> None:
        if self.pidfile:
            try:
                os.unlink(self.pidfile)
            except OSError:
                module_logger.error("Couldn't remove pidfile!")

    def add_monitor(self, name: str, monitor: Monitor) -> None:
        """Add a monitor."""
        self.monitors[name] = monitor

    def update_monitor_config(self, name: str, config_options: dict) -> None:
        """Update the configuration for a monitor."""
        self.monitors[name].__init__(name, config_options)  # type: ignore

    def update_logger_config(self, name: str, config_options: dict) -> None:
        """Update the configration for a logger."""
        self.loggers[name].__init__(config_options)  # type: ignore

    def update_alerter_config(self, name: str, config_options: dict) -> None:
        """Update the configuration for an alerter."""
        self.alerters[name].__init__(config_options)  # type: ignore

    def has_monitor(self, monitor: str) -> bool:
        """Check if a montitor is known."""
        return monitor in self.monitors.keys()

    def has_logger(self, logger: str) -> bool:
        """Check if a logger is known."""
        return logger in self.loggers.keys()

    def has_alerter(self, alerter: str) -> bool:
        """Check if an alerter is known."""
        return alerter in self.alerters.keys()

    def reset_monitors(self) -> None:
        """Clear all all monitors' dependency info back to default."""
        for key in list(self.monitors.keys()):
            self.monitors[key].reset_dependencies()
            self.monitors[key].ran_this_time = False

    def _verify_dependencies(self) -> bool:
        """Check if all monitors have valid dependencies."""
        ok = True
        monitors = self.monitors.keys()
        for key, monitor in self.monitors.items():
            for dependency in monitor.dependencies:
                if dependency not in monitors:
                    module_logger.critical(
                        "Configuration error: dependency %s of monitor %s is not defined or is not enabled!",
                        dependency,
                        key,
                    )
                    ok = False
        return ok

    def verify_alerting(self) -> bool:
        """Sanity check the configuration to see if we have at least an
        alerter, or network logging."""
        sane = True
        if len(self.alerters) == 0:
            for _, logger in self.loggers.items():
                if logger.logger_type == "network":
                    break
            else:
                sane = False
        return sane

    def sort_joblist(self, joblist: List[str]) -> List[str]:
        """Order a list of monitors so that compound monitors are at the end"""
        new_list = []  # type: List[str]
        late_list = []  # type: List[str]
        for monitor in joblist:
            if self.monitors[monitor].monitor_type in ["compound"]:
                late_list.append(monitor)
            else:
                new_list.append(monitor)
        new_list.extend(late_list)
        return new_list

    def _failed_monitors(self) -> List[str]:
        """Return a list of the currently-failed monitors.

        Includes monitors which are disabled."""
        failed = [
            key
            for key, monitor in self.monitors.items()
            if monitor.state() == MonitorState.FAILED or not monitor.enabled
        ]
        return failed

    @staticmethod
    def _run_monitor(monitor: Monitor) -> bool:
        """Run a single monitor."""
        did_run = False
        try:
            if monitor.should_run():
                did_run = True
                monitor.ran_this_time = True
                start_time = time.time()
                monitor.run_test()
                end_time = time.time()
                monitor.last_run_duration = int(end_time - start_time)
            else:
                monitor.record_skip(None)
                module_logger.info("Not run: %s", monitor.name)
        except Exception as exception:
            module_logger.exception(
                "Monitor %s threw exception during run_test()", monitor.name
            )
            monitor.record_fail("Unhandled exception: {}".format(exception))
        if monitor.error_count > 0:
            if monitor.virtual_fail_count() == 0:
                module_logger.warning(
                    "monitor failed but within tolerance: %s (%s)",
                    monitor.name,
                    monitor.last_result,
                )
            else:
                module_logger.error(
                    "monitor failed: %s (%s)",
                    monitor.name,
                    monitor.last_result,
                )
            return False
        if not did_run:
            return False
        module_logger.info("monitor passed: %s", monitor.name)
        return True

    def run_tests(self) -> None:
        """Run the tests for all the monitors."""
        self.reset_monitors()

        joblist = [k for (k, v) in self.monitors.items() if v.enabled]
        while joblist:
            new_joblist, skiplist = self._prepare_lists(joblist)
            joblist = self.sort_joblist(joblist)
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=self._max_workers
            ) as executor:
                future_to_monitor = {}
                for monitor in joblist:
                    if monitor in new_joblist or monitor in skiplist:
                        module_logger.debug(
                            "Skipping monitor %s because it's in the new job list or the skiplist",
                            monitor,
                        )
                        continue
                    module_logger.debug("Trying monitor: %s", monitor)
                    future_to_monitor[
                        executor.submit(self._run_monitor, self.monitors[monitor])
                    ] = monitor
                if len(future_to_monitor) == 0:
                    module_logger.error("No more monitors are runnable!")
                    return
                for future in concurrent.futures.as_completed(future_to_monitor):
                    monitor = future_to_monitor[future]
                    try:
                        if future.result():
                            for monitor2 in joblist:
                                self.monitors[monitor2].dependency_succeeded(monitor)
                    except Exception:
                        module_logger.exception(
                            "Exception for monitor %s during thread execution", monitor
                        )
            joblist = copy.copy(new_joblist)

    def _prepare_lists(self, joblist: List[str]) -> Tuple[List[str], List[str]]:
        failed = self._failed_monitors()
        module_logger.debug(
            "Starting loop with joblist %s and failed list %s",
            ", ".join(joblist),
            ", ".join(failed),
        )
        new_joblist = []  # List[str]
        skiplist = []  # List[str]
        for monitor in joblist:
            if monitor in failed:
                module_logger.error(
                    "Received a failed logger in the joblist: %s", monitor
                )
                continue
            if self.monitors[monitor].monitor_type == "compound":
                # special case handling for compound monitors
                compound_monitor = cast(CompoundMonitor, self.monitors[monitor])
                needed_monitors = set(compound_monitor.monitors)
                remaining_monitors = needed_monitors & set(joblist)
                if remaining_monitors:
                    module_logger.debug(
                        "Added compound monitor %s to new joblist due to outstanding deps %s",
                        monitor,
                        remaining_monitors,
                    )
                    new_joblist.append(monitor)
                    continue
            if self.monitors[monitor].remaining_dependencies:
                # this monitor has outstanding deps, put it on the new joblist for next loop
                failed_deps = set(
                    self.monitors[monitor].remaining_dependencies
                ).intersection(failed)
                if len(failed_deps) > 0:
                    module_logger.warning(
                        "Monitor %s has failed dependencies %s, skipping",
                        monitor,
                        ", ".join(failed_deps),
                    )
                    self.monitors[monitor].record_skip(", ".join(failed_deps))
                    skiplist.append(monitor)
                else:
                    new_joblist.append(monitor)
                    module_logger.debug(
                        "Added %s to new joblist due to outstanding deps %s",
                        monitor,
                        ", ".join(self.monitors[monitor].remaining_dependencies),
                    )
                continue
        return (new_joblist, skiplist)

    def log_result(self, logger: Logger) -> None:
        """Use the given logger object to log our state."""
        logger.check_dependencies(self.failed + self.still_failing + self.skipped)
        with logger:
            for key, monitor in self.monitors.items():
                if not check_group_match(monitor.group, logger.groups):
                    module_logger.debug(
                        "not logging for %s due to group mismatch (monitor in group %s, "
                        "logger has groups %s",
                        key,
                        monitor.group,
                        logger.groups,
                    )
                    continue
                if logger.heartbeat and not monitor.ran_this_time:
                    module_logger.debug(
                        "not logging for %s as this logger is in heartbeat mode and"
                        " the monitor did not run this loop",
                        key,
                    )
                    continue
                logger.save_result2(key, monitor)

            # need to work on a copy here to prevent the dicts changing under us
            # during the loop, as remote instances can connect and update our data
            # unpredictably
            for host_monitors in self.remote_monitors.copy().values():
                for name, monitor in host_monitors.copy().items():
                    try:
                        if check_group_match(monitor.group, logger.groups):
                            logger.save_result2(name, monitor)
                        else:
                            module_logger.debug(
                                "not logging for %s due to group mismatch (monitor in group %s, "
                                "logger has groups %s",
                                name,
                                monitor.group,
                                logger.groups,
                            )
                    except Exception:  # pragma: no cover
                        module_logger.exception(
                            "exception while logging remote monitors (offending monitor: %s)",
                            name,
                        )

    def do_alert(self, alerter: Alerter) -> None:
        """Use the given alerter object to send an alert, if needed."""
        alerter.check_dependencies(self.failed + self.still_failing + self.skipped)
        for name, this_monitor in list(self.monitors.items()):
            # Don't generate alerts for monitors which want it done remotely
            if this_monitor.remote_alerting:
                module_logger.debug(
                    "skipping alert for monitor %s as it wants remote alerting", name
                )
                continue
            try:
                if this_monitor.notify:
                    alerter.send_alert(name, this_monitor)
                else:
                    module_logger.warning("monitor %s has notifications disabled", name)
            except Exception:  # pragma: no cover
                module_logger.exception("exception caught while alerting for %s", name)
        for host_monitors in self.remote_monitors.copy().values():
            for name, monitor in host_monitors.copy().items():
                try:
                    if monitor.remote_alerting:
                        alerter.send_alert(name, monitor)
                    else:
                        module_logger.debug(
                            "not alerting for monitor %s as it doesn't want remote alerts",
                            name,
                        )
                except Exception:  # pragma: no cover
                    module_logger.exception(
                        "exception caught while alerting for remote monitor %s", name
                    )

    def count_monitors(self) -> int:
        """Gets the number of monitors we have defined."""
        return len(self.monitors)

    def add_alerter(self, name: str, alerter: Alerter) -> None:
        """Add an alerter."""
        if isinstance(alerter, Alerter):
            self.alerters[name] = alerter
        else:
            module_logger.critical(
                "Failed to add alerter because it is not the right type"
            )

    def add_logger(self, name: str, logger: Logger) -> None:
        """Add a logger."""
        if isinstance(logger, Logger):
            self.loggers[name] = logger
        else:
            module_logger.critical(
                "Failed to add logger because it is not the right type"
            )

    def prune_monitors(self, retain: List[str]) -> None:
        """Remove monitors which are in our list but not in the list passed to us.

        Used to tidy up after a config reload (which may have removed monitors)"""
        delete_list = []
        for monitor in self.monitors:
            if monitor not in retain:
                module_logger.info("Removing monitor %s", monitor)
                delete_list.append(monitor)
        for monitor in delete_list:
            del self.monitors[monitor]
        if not self._verify_dependencies():
            module_logger.critical(
                "Broken dependencies after pruning monitors, aborting!"
            )
            sys.exit(1)

    def prune_alerters(self, retain: List[str]) -> None:
        """Remove alerters which are in our list but not in the list passed to us.

        Used to tidy up after a config reload (which may have removed alerters)"""
        delete_list = []
        for alerter in self.alerters:
            if alerter not in retain:
                module_logger.info("Removing alerter %s", alerter)
                delete_list.append(alerter)
        for alerter in delete_list:
            del self.alerters[alerter]

    def prune_loggers(self, retain: List[str]) -> None:
        """Remove loggers which are in our list but not in the list passed to us.

        Used to tidy up after a config reload (which may have removed logger)"""
        delete_list = []
        for logger in self.loggers:
            if logger not in retain:
                module_logger.info("Removing logger %s", logger)
                delete_list.append(logger)
        for logger in delete_list:
            del self.loggers[logger]

    def do_alerts(self) -> None:
        """Run the alert process for each alerter."""
        for alerter in self.alerters.values():
            self.do_alert(alerter)

    def do_recovery(self) -> None:
        """Attempt recovery for each monitor."""
        for monitor in self.monitors.values():
            monitor.attempt_recover()

    def do_recovered(self) -> None:
        """Run the recovered action for each monitor."""
        for monitor in self.monitors.values():
            monitor.run_recovered()

    def hup_loggers(self) -> None:
        """Inform each logger they need to HUP."""
        for logger in self.loggers.values():
            logger.hup()

    def do_logs(self) -> None:
        """Log result for each logger."""
        for logger in self.loggers.values():
            self.log_result(logger)

    def update_remote_monitor(self, data: Dict[str, dict], hostname: str) -> None:
        """Process a list of monitors received from a remote host."""
        seen_monitors = []  # type: List[str]
        if hostname not in self.remote_monitors:
            self.remote_monitors[hostname] = {}
        for name, state in data.items():
            module_logger.info(
                "updating remote monitor %s from host %s", name, hostname
            )
            if isinstance(state, dict):
                try:
                    remote_monitor = get_monitor_class(
                        state["cls_type"]
                    ).from_python_dict(state["data"])
                    self.remote_monitors[hostname][name] = remote_monitor
                    seen_monitors.append(name)
                except KeyError:
                    module_logger.exception(
                        "Could not add remote monitor from host %s; "
                        "possibly a monitor type we don't know?",
                        hostname,
                    )
            else:
                module_logger.critical(
                    "Could not deserialize state of monitor %s. "
                    "If the remote host uses an old version of "
                    "simplemonitor, you need to upgrade.",
                    name,
                )
        self._trim_remote_monitors(hostname, seen_monitors)

    def _trim_remote_monitors(self, hostname: str, seen_monitors: List[str]) -> None:
        """Remove remote monitors for a host which aren't in the given list."""
        host_monitors = self.remote_monitors[hostname]
        forget_monitors = []
        for name in host_monitors.keys():
            if name not in seen_monitors:
                module_logger.info(
                    "forgetting remote monitor %s from host %s", name, hostname
                )
                forget_monitors.append(name)
        for name in forget_monitors:
            del self.remote_monitors[hostname][name]

    @property
    def remote_hosts(self) -> dict[str, RemoteHost]:
        return self._remote_hosts

    def upsert_remote_host(self, host: str, last_seen: datetime, address: str) -> None:
        self._remote_hosts[host] = {"last_seen": last_seen, "address": address}

    def run_loop(self) -> None:
        """Run the complete monitor loop once."""
        module_logger.debug("Running tests")
        self.run_tests()
        module_logger.debug("Running recovery")
        self.do_recovery()
        self.do_recovered()
        module_logger.debug("Running alerts")
        self.do_alerts()
        module_logger.debug("Running logs")
        self.do_logs()
        module_logger.debug("Loop complete")

    def run(self) -> None:
        self._create_pid_file()
        self._start_network_thread()
        module_logger.info(
            "=== Starting... (loop runs every %ds) Hit ^C to stop", self.interval
        )
        loop = True
        loops = self._max_loops
        heartbeat = True
        while loop:
            try:
                if loops > 0:
                    loops -= 1
                    if loops == 0:
                        module_logger.warning(
                            "Ran out of loop counter, will stop after this one"
                        )
                        loop = False
                if self._need_hup or self._check_hup_file():
                    try:
                        module_logger.warning("Reloading configuration")
                        self._load_config()
                        self._start_network_thread()
                        self.hup_loggers()
                        self._need_hup = False
                    except Exception:
                        module_logger.exception("Error while reloading configuration")
                        sys.exit(1)
                self.run_loop()

                if (
                    module_logger.level in ["error", "critical", "warn"]
                    and self.heartbeat
                ):
                    heartbeat = not heartbeat
                    if heartbeat:
                        sys.stdout.write(".")
                        sys.stdout.flush()
            except KeyboardInterrupt:
                module_logger.info("Received ^C")
                loop = False
            except Exception:
                module_logger.exception("Caught unhandled exception during main loop")
            if loop and self._network:
                if (
                    self._remote_listening_thread
                    and not self._remote_listening_thread.is_alive()
                ):
                    module_logger.error("Listener thread died :(")
                    self._start_network_thread()
            if self.one_shot:
                break

            try:
                if loop:
                    time.sleep(self.interval)
            except Exception:
                module_logger.info("Quitting")
                loop = False

        self._remove_pid_file()



# type: ignore
import unittest

from simplemonitor.Monitors import host


class TestHostMonitors(unittest.TestCase):
    safe_config = {"partition": "/", "limit": "10G"}

    one_KB = 1024
    one_MB = one_KB * 1024
    one_GB = one_MB * 1024
    one_TB = one_GB * 1024

    def test_DiskSpace_brokenConfigOne(self):
        config_options = {}
        with self.assertRaises(ValueError):
            host.MonitorDiskSpace("test", config_options)

    def test_DiskSpace_brokenConfigTwo(self):
        config_options = {"partition": "/"}
        with self.assertRaises(ValueError):
            host.MonitorDiskSpace("test", config_options)

    def test_DiskSpace_brokenConfigThree(self):
        config_options = {"partition": "/", "limit": "moo"}
        with self.assertRaises(ValueError):
            host.MonitorDiskSpace("test", config_options)

    def test_DiskSpace_correctConfig(self):
        m = host.MonitorDiskSpace("test", self.safe_config)
        self.assertIsInstance(m, host.MonitorDiskSpace)

    def test_DiskSpace_meta(self):
        m = host.MonitorDiskSpace("test", self.safe_config)

        self.assertEqual(
            m.describe(),
            "Checking for at least 10.00GiB free space on /",
            "Failed to verify description",
        )

        self.assertTupleEqual(
            m.get_params(), (10 * self.one_GB, "/"), "Failed to verify params"
        )

    def test_DiskSpace_free(self):
        # Hopefully our test machine has at least 1 byte free on /
        config_options = {"partition": "/", "limit": "1"}
        m = host.MonitorDiskSpace("test", config_options)
        m.run_test()
        self.assertTrue(m.test_success())

        # and hopefully it has a sensible-sized root partition
        config_options = {"partition": "/", "limit": "100000G"}
        m = host.MonitorDiskSpace("test", config_options)
        m.run_test()
        self.assertFalse(m.test_success())

    def test_DiskSpace_invalid_partition(self):
        config_options = {"partition": "moo", "limit": "1"}
        m = host.MonitorDiskSpace("test", config_options)
        m.run_test()
        self.assertFalse(m.test_success(), "Monitor did not fail")
        self.assertRegex(
            m.last_result,
            "Couldn't get free disk space",
            "Monitor did not report error correctly",
        )

    def test_DiskSpace_get_params(self):
        config_options = {"partition": "/", "limit": "1"}
        m = host.MonitorDiskSpace("test", config_options)
        self.assertTupleEqual(m.get_params(), (1, "/"))

    def test_Filestat_get_params(self):
        config_options = {"maxage": "10", "minsize": "20", "filename": "/test"}
        m = host.MonitorFileStat("test", config_options)
        self.assertTupleEqual(m.get_params(), ("/test", 20, 10))

    def test_Command_get_params(self):
        config_options = {"command": "ls /", "result_regexp": "moo", "result_max": "10"}
        m = host.MonitorCommand("test", config_options)
        self.assertTupleEqual(m.get_params(), (["ls", "/"], "moo", None, False))

        config_options = {"command": "ls /", "result_max": "10"}
        m = host.MonitorCommand("test", config_options)
        self.assertTupleEqual(m.get_params(), (["ls", "/"], "", 10, False))

        config_options = {"command": "ls /", "result_max": "10", "show_output": True}
        m = host.MonitorCommand("test", config_options)
        self.assertTupleEqual(m.get_params(), (["ls", "/"], "", 10, True))


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import Mock, patch

from requests import Response
from requests.auth import HTTPBasicAuth

from simplemonitor.Monitors import MonitorHTTP
from simplemonitor.util import MonitorState


class TestMonitorHTTP(unittest.TestCase):
    def test_get_ok_default_params(self):
        response_mock = Mock(spec=Response)
        response_mock.status_code = 200

        monitor = MonitorHTTP(
            name="test_http_monitor",
            config_options={
                "url": "http://example.com",
                "urgent": 0,
                "tolerance": 1,
                "remote_alert": 1,
            },
        )

        with patch("requests.request", return_value=response_mock) as mock:
            monitor.run_test()

        result = monitor.get_result()
        state = monitor.state()

        mock.assert_called_once_with(
            "GET",
            "http://example.com",
            headers=None,
            timeout=5,
            verify=True,
            allow_redirects=True,
            auth=None,
            cert=None,
            data=None,
            json=None,
        )
        self.assertEqual(state, MonitorState.OK)
        self.assertIn("200", result)

    def test_get_ok_non_default_params(self):
        response_mock = Mock(spec=Response)
        response_mock.status_code = 200

        monitor = MonitorHTTP(
            name="test_http_monitor",
            config_options={
                "url": "http://example.com",
                "urgent": 0,
                "tolerance": 1,
                "remote_alert": 1,
                "verify_hostname": False,
                "allow_redirects": False,
                "username": "test",
                "password": "pass",
                "headers": '{"test": "header"}',
                "timeout": 10,
            },
        )

        with patch("requests.request", return_value=response_mock) as mock:
            monitor.run_test()

        result = monitor.get_result()
        state = monitor.state()

        mock.assert_called_once_with(
            "GET",
            "http://example.com",
            headers={"test": "header"},
            timeout=10,
            verify=False,
            allow_redirects=False,
            auth=HTTPBasicAuth("test", "pass"),
            cert=None,
            data=None,
            json=None,
        )
        self.assertEqual(state, MonitorState.OK)
        self.assertIn("200", result)
        pass



"""
SimpleMonitor alerts via Slack webhooks
"""

from typing import Any, Dict, cast

import requests

from ..Monitors.monitor import Monitor
from ..util import format_datetime
from .alerter import Alerter, AlertType, register


@register
class SlackAlerter(Alerter):
    """Send alerts to a Slack webhook"""

    alerter_type = "slack"

    channel = None
    username = None

    def __init__(self, config_options: dict) -> None:
        super().__init__(config_options)
        self.url = cast(
            str, self.get_config_option("url", required=True, allow_empty=False)
        )

        self.channel = cast(str, self.get_config_option("channel"))
        self.username = cast(str, self.get_config_option("username"))
        self.timeout = cast(
            int, self.get_config_option("timeout", required_type="int", default=5)
        )

    def send_alert(self, name: str, monitor: Monitor) -> None:
        """Send the message"""
        alert_type = self.should_alert(monitor)
        downtime = monitor.get_downtime()

        message_json = {}  # type: Dict[str, Any]
        if self.channel is not None:
            message_json = {"channel": self.channel}
        elif self.username is not None:
            message_json = {"username": self.username}
        else:
            message_json = {}

        message_json["attachments"] = [{}]

        if alert_type == AlertType.NONE:
            return
        if alert_type == AlertType.FAILURE:
            message_json["text"] = f"Monitor {name} failed!"
            message_json["attachments"][0]["color"] = "danger"
            fields = [
                {
                    "title": "Failed at",
                    "value": format_datetime(monitor.first_failure_time()),
                    "short": True,
                },
                {"title": "Downtime", "value": str(downtime), "short": True},
                {
                    "title": "Virtual failure count",
                    "value": monitor.virtual_fail_count(),
                    "short": True,
                },
                {"title": "Host", "value": self.hostname, "short": True},
                {"title": "Additional info", "value": monitor.get_result()},
                {"title": "Description", "value": monitor.describe()},
            ]

            if monitor.failure_doc:
                fields.append(
                    {
                        "title": "Documentation",
                        "value": monitor.failure_doc,
                    }
                )

            try:
                if monitor.recover_info != "":
                    fields.append(
                        {
                            "title": "Recovery info",
                            "value": f"Recovery info: {monitor.recover_info}",
                        }
                    )
                    message_json["attachments"][0]["color"] = "warning"
            except AttributeError:
                pass
            message_json["attachments"][0]["fields"] = fields

        elif alert_type == AlertType.SUCCESS:
            message_json["text"] = f"Monitor {name} succeeded."
            fields = [
                {
                    "title": "Failed at",
                    "value": format_datetime(monitor.first_failure_time()),
                    "short": True,
                },
                {"title": "Downtime", "value": str(downtime), "short": True},
                {"title": "Host", "value": self.hostname, "short": True},
                {"title": "Description", "value": monitor.describe()},
            ]
            message_json["attachments"][0]["color"] = "good"
            message_json["attachments"][0]["fields"] = fields

        else:
            self.alerter_logger.error("unknown alert type %s", alert_type)
            return

        if not self._dry_run:
            try:
                response = requests.post(
                    self.url, json=message_json, timeout=self.timeout
                )
                if not response.status_code == 200:
                    self.alerter_logger.error(
                        "POST to slack webhook failed: %s", response
                    )
            except requests.exceptions.RequestException:
                self.alerter_logger.exception("Failed to post to slack webhook")
        else:
            self.alerter_logger.info(
                "dry_run: would send slack: %s", message_json.__repr__()
            )

    def _describe_action(self) -> str:
        target = ""
        if self.channel:
            target = self.channel
        elif self.username:
            target = self.username
        if target:
            return f"posting to {target} on Slack"
        return "posting to Slack"



"""
Loggers for SimpleMonitor
"""

from .db import DBFullLogger, DBStatusLogger
from .file import FileLogger
from .mqtt import MQTTLogger
from .network import Listener, NetworkLogger
from .seq import SeqLogger

__all__ = [
    "DBFullLogger",
    "DBStatusLogger",
    "FileLogger",
    "Listener",
    "MQTTLogger",
    "NetworkLogger",
    "SeqLogger",
]



"""
SimpleMonitor logging with sqlite
"""

try:
    import sqlite3

    SQLLITE_AVAILABLE = True
except ImportError:
    SQLLITE_AVAILABLE = False

import time
from socket import gethostname

from ..Monitors.monitor import Monitor
from .logger import Logger, register

CREATE_SQL = [
    """
-- sqlite3 schema for monitor.db
-- version 1
CREATE TABLE IF NOT EXISTS results(
result_id integer primary key,
monitor_host varchar(50),
monitor_name varchar(50),
monitor_type varchar(50),
monitor_params varchar(100),
monitor_result int,
timestamp int,
monitor_info varchar(255));

CREATE TABLE IF NOT EXISTS status (
monitor_host varchar(50),
monitor_name varchar(50),
monitor_result int,
monitor_info varchar(255));

CREATE TABLE IF NOT EXISTS monitor_schema (
    k varchar(50) primary key,
    v varchar(255)
);

INSERT OR IGNORE INTO monitor_schema (k, v) VALUES ('monitor_schema_version', 1)
"""
]


class DBLogger(Logger):
    """Abstract class which uses a sqlite3 backend."""

    hostname = gethostname()
    connected = False

    def __init__(self, config_options: dict) -> None:
        """Open the database connection."""
        super().__init__(config_options)
        if not SQLLITE_AVAILABLE:
            self.logger_logger.critical(
                "Python sqlite support not available. Maybe you need to install "
                "sqlite-devel or libsqlite3-dev, or similar?"
            )
            self.connected = False
            return
        self.db_path = self.get_config_option(
            "db_path", required=True, allow_empty=False
        )

        self.db_handle = sqlite3.connect(self.db_path, isolation_level=None)
        self.db_handle.row_factory = sqlite3.Row
        self.connected = True
        self.check_schema()

    def check_schema(self) -> None:
        """Create tables if needed, and check the schema."""
        self.db_handle.executescript(CREATE_SQL[0])
        cursor = self.db_handle.cursor()
        current_schema = None
        expected_schema = len(CREATE_SQL)
        for row in cursor.execute(
            "SELECT v AS value FROM monitor_schema where k = 'monitor_schema_version'"
        ):
            current_schema = int(row["value"])
        if current_schema is None:
            self.logger_logger.error(
                "Could not check current schema version! Expect weirdness."
            )
            return
        if current_schema < expected_schema:
            self.logger_logger.warning(
                "Schema for %s is out of date: current is %d, latest is %d.",
                self.db_path,
                current_schema,
                expected_schema,
            )
            self.roll_schema_forward(current_schema)
        elif current_schema > expected_schema:
            self.logger_logger.critical(
                "Schema for %s is newer than this code! Cannot use this database file.",
                self.db_path,
            )
            self.connected = False
        else:
            self.logger_logger.debug("Schema for %s is current", self.db_path)

    def roll_schema_forward(self, start: int) -> None:
        """Update the db schema"""
        for sql in CREATE_SQL[start:]:
            self.logger_logger.info("Applying SQL schema update")
            self.logger_logger.debug(sql)
            try:
                self.db_handle.executescript(sql)
            except Exception:
                self.logger_logger.exception("Failed to apply schema update")
                self.logger_logger.critical(
                    "Cannot use this DB logger until schema is fixed!"
                )
                self.connected = False

    def save_result2(self, name: str, monitor: Monitor) -> None:
        raise NotImplementedError


@register
class DBFullLogger(DBLogger):
    """Logs results to a sqlite3 db."""

    logger_type = "db"

    def save_result(
        self,
        monitor_name: str,
        monitor_type: str,
        monitor_params: str,
        monitor_result: int,
        monitor_info: str,
        hostname: str = "",
    ) -> None:
        """Write to the database."""
        if not self.connected:
            self.logger_logger.warning("cannot send results, a dependency failed")
            return
        sql = (
            "INSERT INTO results (result_id, monitor_host, monitor_name, monitor_type, "
            "monitor_params, monitor_result, timestamp, monitor_info) "
            "VALUES (null, ?, ?, ?, ?, ?, ?, ?)"
        )

        cursor = self.db_handle.cursor()

        join_string = ":"
        timestamp = int(time.time())
        if hostname == "":
            hostname = self.hostname

        params = (
            hostname,
            monitor_name,
            monitor_type,
            join_string.join([str(x) for x in monitor_params]),
            monitor_result,
            timestamp,
            monitor_info,
        )
        try:
            cursor.execute(sql, params)
        except sqlite3.OperationalError:
            self.logger_logger.exception("sqlite failed to write to database")

    def save_result2(self, name: str, monitor: Monitor) -> None:
        """new interface."""
        if monitor.test_success():
            result = 1
        else:
            result = 0
        self.save_result(
            name,
            monitor.monitor_type,
            str(monitor.get_params()),
            result,
            monitor.describe(),
        )

    def describe(self) -> str:
        return "Logging results to {0}".format(self.db_path)


@register
class DBStatusLogger(DBLogger):
    """Maintains status snapshot in db."""

    logger_type = "dbstatus"

    def _save_result(
        self,
        monitor_name: str,
        monitor_type: str,
        monitor_params: str,
        monitor_result: int,
        monitor_info: str,
        hostname: str = "",
    ) -> None:
        """Original implementation for saving a result"""
        if hostname == "":
            hostname = self.hostname
        cursor = self.db_handle.cursor()
        try:
            cursor.execute(
                "DELETE FROM status WHERE monitor_host = ? AND monitor_name = ?",
                (self.hostname, monitor_name),
            )
            cursor.execute(
                (
                    "REPLACE INTO status (monitor_host, monitor_name, "
                    "monitor_result, monitor_info) "
                    "VALUES (?, ?, ?, ?)"
                ),
                (hostname, monitor_name, monitor_result, monitor_info),
            )
        except sqlite3.OperationalError:
            self.logger_logger.exception("sqlite failed to write to database")

    def save_result2(self, name: str, monitor: Monitor) -> None:
        """Record a result"""
        if monitor.test_success():
            result = 1
        else:
            result = 0
        self._save_result(
            name,
            monitor.monitor_type,
            str(monitor.get_params()),
            result,
            monitor.describe(),
        )

    def describe(self) -> str:
        return f"Logging status to {self.db_path}"



# coding=utf-8
"""A (fairly) simple host/service monitor."""

import argparse
import logging
import os
import sys

from .const import EXIT_CODE_CONFIG_FAILED
from .simplemonitor import SimpleMonitor
from .version import VERSION

try:
    import colorlog
except ImportError:
    pass


main_logger = logging.getLogger("simplemonitor")


def main() -> None:
    r"""This is where it happens \o/"""

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--version", action="version", version="%(prog)s {}".format(VERSION)
    )
    output_group = parser.add_argument_group(title="Output controls")
    testing_group = parser.add_argument_group(title="Test and debug tools")
    output_group.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        dest="verbose",
        default=False,
        help="Alias for --log-level=info",
    )
    output_group.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        dest="quiet",
        default=False,
        help="Alias for --log-level=critical",
    )
    testing_group.add_argument(
        "-t",
        "--test",
        action="store_true",
        dest="test",
        default=False,
        help="Test config and exit",
    )
    parser.add_argument(
        "-p", "--pidfile", dest="pidfile", default=None, help="Write PID into this file"
    )
    parser.add_argument(
        "-N",
        "--no-network",
        dest="no_network",
        default=False,
        action="store_true",
        help="Disable network listening socket (if enabled in config)",
    )
    output_group.add_argument(
        "-d",
        "--debug",
        dest="debug",
        default=False,
        action="store_true",
        help="Alias for --log-level=debug",
    )
    parser.add_argument(
        "-f",
        "--config",
        dest="config",
        default="monitor.ini",
        help=(
            "configuration file (this is the main config; "
            "you also need monitors.ini (default filename))"
        ),
    )
    parser.add_argument(
        "-j",
        "--threads",
        dest="threads",
        default=os.cpu_count(),  # default used by the library anyway
        type=int,
        help=(
            f"number of threads to run for checking monitors (default (cpus): {os.cpu_count()})"
        ),
    )
    output_group.add_argument(
        "-H",
        "--no-heartbeat",
        action="store_true",
        dest="no_heartbeat",
        default=False,
        help="Omit printing the '.' character when running checks",
    )
    testing_group.add_argument(
        "-1",
        "--one-shot",
        action="store_true",
        dest="one_shot",
        default=False,
        help=(
            "Run the monitors once only, without alerting. Require monitors without "
            '"fail" in the name to succeed. Exit zero or non-zero accordingly'
        ),
    )
    testing_group.add_argument(
        "--loops",
        dest="loops",
        default=-1,
        type=int,
        help="Number of iterations to run before exiting",
    )
    output_group.add_argument(
        "-l",
        "--log-level",
        dest="loglevel",
        default="warn",
        help="Log level: critical, error, warn, info, debug",
    )
    output_group.add_argument(
        "-C",
        "--no-colour",
        "--no-color",
        action="store_true",
        dest="no_colour",
        default=False,
        help="Do not colourise log output",
    )
    output_group.add_argument(
        "--no-timestamps",
        action="store_true",
        dest="no_timestamps",
        default=False,
        help="Do not prefix log output with timestamps",
    )
    testing_group.add_argument(
        "--dump-known-resources",
        action="store_true",
        dest="dump_resources",
        default=False,
        help="Print out loaded Monitor, Alerter and Logger types",
    )

    options = parser.parse_args()

    if options.dump_resources:
        import pprint

        import simplemonitor.Alerters.alerter as alerter
        import simplemonitor.Loggers.logger as logger
        import simplemonitor.Monitors.monitor as monitor

        print("Monitors:")
        pprint.pprint(sorted(monitor.all_types()), compact=True)
        print("Loggers:")
        pprint.pprint(sorted(logger.all_types()), compact=True)
        print("Alerters:")
        pprint.pprint(sorted(alerter.all_types()), compact=True)
        sys.exit(0)

    if options.quiet:
        options.loglevel = "critical"

    if options.verbose:
        options.loglevel = "info"

    if options.debug:
        options.loglevel = "debug"

    if options.no_timestamps:
        logging_timestamp = ""
    else:
        logging_timestamp = "%(asctime)s "

    try:
        log_level = getattr(logging, options.loglevel.upper())
    except AttributeError:
        print("Log level {0} is unknown".format(options.loglevel))
        sys.exit(1)

    log_datefmt = "%Y-%m-%d %H:%M:%S"
    log_plain_format = logging_timestamp + "%(levelname)8s (%(name)s) %(message)s"
    if not options.no_colour:
        try:
            handler = colorlog.StreamHandler()
            handler.setFormatter(
                colorlog.ColoredFormatter(
                    logging_timestamp
                    + "%(log_color)s%(levelname)8s%(reset)s (%(name)s) %(message)s",
                    datefmt="%Y-%m-%d %H:%M:%S",
                )
            )
            main_logger.addHandler(handler)
        except NameError:
            logging.basicConfig(format=log_plain_format, datefmt=log_datefmt)
            main_logger.error("Could not enable colorlog")
    else:
        logging.basicConfig(format=log_plain_format, datefmt=log_datefmt)

    main_logger.setLevel(log_level)

    if not options.quiet:
        main_logger.info("=== SimpleMonitor v%s", VERSION)
        main_logger.info("Loading main config from %s", options.config)

    m = SimpleMonitor(
        config_file=options.config,
        no_network=options.no_network,
        max_loops=options.loops,
        heartbeat=not options.no_heartbeat,
        one_shot=options.one_shot,
        max_workers=options.threads,
    )

    if m.config_ok:
        if options.test:
            main_logger.warning("Config test complete. Exiting.")
            sys.exit(0)
    else:
        main_logger.error("Configuration not valid")
        sys.exit(EXIT_CODE_CONFIG_FAILED)

    if options.one_shot:
        main_logger.warning(
            "One-shot mode: expecting monitors without 'fail' in the name to succeed, "
            "and with to fail. Will exit zero or non-zero accordingly."
        )

    m.run()

    main_logger.info("Finished.")

    if options.one_shot:  # pragma: no cover
        ok = True
        print("\n--> One-shot results:")
        tail_info = []
        for this_monitor in sorted(m.monitors.keys()):
            if "fail" in this_monitor:
                if m.monitors[this_monitor].error_count == 0:
                    tail_info.append(
                        "    Monitor {0} should have failed".format(this_monitor)
                    )
                    tail_info.append(
                        "        {}".format(m.monitors[this_monitor].last_result)
                    )
                    ok = False
                else:
                    print("    Monitor {0} was ok (failed)".format(this_monitor))
            elif "skip" in this_monitor:
                if m.monitors[this_monitor].skipped():
                    print("    Monitor {0} was ok (skipped)".format(this_monitor))
                else:
                    tail_info.append(
                        "    Monitor {0} should have been skipped".format(this_monitor)
                    )
                    ok = False
            else:
                if m.monitors[this_monitor].error_count > 0:
                    tail_info.append(
                        "    Monitor {0} failed and shouldn't have: {1}".format(
                            this_monitor, m.monitors[this_monitor].last_result
                        )
                    )
                    ok = False
                    tail_info.append(
                        "        {}".format(m.monitors[this_monitor].last_result)
                    )
                else:
                    print("    Monitor {0} was ok".format(this_monitor))
        if len(tail_info):
            print()
            for line in tail_info:
                print(line)
        if not ok:
            print("Not all non-'fail' succeeded, or not all 'fail' monitors failed.")
            sys.exit(1)

    logging.shutdown()


if __name__ == "__main__":
    main()



# type: ignore
import os
import unittest

from simplemonitor.util import envconfig

os.environ["TEST_VALUE"] = "test1"


class TestEnvConfig(unittest.TestCase):
    def test_EnvConfig(self):
        config = envconfig.EnvironmentAwareConfigParser()
        config.read("tests/monitor-env.ini")
        self.assertEqual(config.get("monitor", "monitors"), "tests/monitors-test1.ini")

    def test_EnvConfigKey(self):
        config = envconfig.EnvironmentAwareConfigParser()
        config.read("tests/monitor-env.ini")
        self.assertEqual(config.get("monitor-test1", "monitors"), "hello")



"""
Alerters for SimpleMonitor
"""

from .bulksms import BulkSMSAlerter
from .execute import ExecuteAlerter
from .fortysixelks import FortySixElksAlerter
from .healthchecks import HealthchecksAlerter
from .mail import EMailAlerter
from .nc import NotificationCenterAlerter
from .nextcloud_notification import NextcloudNotificationAlerter
from .ntfy import NtfyAlerter
from .pushbullet import PushbulletAlerter
from .pushover import PushoverAlerter
from .ses import SESAlerter
from .slack import SlackAlerter
from .sms77 import SMS77Alerter
from .sns import SNSAlerter
from .syslogger import SyslogAlerter
from .telegram import TelegramAlerter
from .twilio import TwilioSMSAlerter

__all__ = [
    "BulkSMSAlerter",
    "EMailAlerter",
    "ExecuteAlerter",
    "FortySixElksAlerter",
    "HealthchecksAlerter",
    "NotificationCenterAlerter",
    "NextcloudNotificationAlerter",
    "NtfyAlerter",
    "PushbulletAlerter",
    "PushoverAlerter",
    "SESAlerter",
    "SlackAlerter",
    "SMS77Alerter",
    "SNSAlerter",
    "SyslogAlerter",
    "TelegramAlerter",
    "TwilioSMSAlerter",
]
